import datetime
import os

from typing import List, Optional, Tuple

import aiofiles

from haupt.streams.tasks.logs import (
    content_to_logs,
    download_logs_file,
    download_tmp_logs,
)
from polyaxon._fs.async_manager import list_files
from polyaxon._fs.types import FSSystem
from polyaxon._k8s.logging.async_monitor import query_k8s_operation_logs
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from traceml.logging import V1Log


async def get_logs_files(fs: FSSystem, store_path: str, run_uuid: str) -> List[str]:
    files = await list_files(
        fs=fs, store_path=store_path, subpath="{}/plxlogs".format(run_uuid)
    )
    if not files["files"]:
        return []
    return sorted([f for f in files["files"].keys()])


async def get_next_file(
    files: List[str], last_file: Optional[str] = None
) -> Optional[str]:
    if not files:
        return None

    if not last_file:
        return files[0]

    i = 0
    for i, f in enumerate(files):
        if f == last_file:
            break
    i += 1
    if i >= len(files):
        return None

    return files[i]


async def read_logs_file(logs_path) -> List[V1Log]:
    if not logs_path or not os.path.exists(logs_path):
        return []

    async with aiofiles.open(logs_path, mode="r") as f:
        content = await f.read()
        return await content_to_logs(content, logs_path)


async def get_archived_operation_logs(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    last_file: Optional[str],
    check_cache: bool = True,
) -> Tuple[List[V1Log], Optional[str], List[str]]:
    files = await get_logs_files(fs=fs, store_path=store_path, run_uuid=run_uuid)
    logs = []
    last_file = await get_next_file(files=files, last_file=last_file)
    if not last_file:
        return logs, last_file, files

    logs = await download_logs_file(
        fs=fs,
        store_path=store_path,
        run_uuid=run_uuid,
        last_file=last_file,
        check_cache=check_cache,
    )

    return logs, last_file, files


async def get_tmp_operation_logs(
    fs: FSSystem, store_path: str, run_uuid: str, last_time: Optional[datetime.datetime]
) -> Tuple[List[V1Log], Optional[datetime.datetime]]:
    logs = []

    tmp_logs = await download_tmp_logs(fs=fs, store_path=store_path, run_uuid=run_uuid)

    if not tmp_logs or not os.path.exists(tmp_logs):
        return logs, None

    tmp_log_files = os.listdir(tmp_logs)
    if not tmp_log_files:
        return logs, None

    for tmp_file in tmp_log_files:
        logs_path = os.path.join(tmp_logs, tmp_file)
        logs += await read_logs_file(logs_path)

    if last_time:
        logs = [l for l in logs if l.timestamp > last_time]
    if logs:
        logs = sorted(logs, key=lambda x: x.timestamp)
        last_time = logs[-1].timestamp
    return [l.to_dict() for l in logs], last_time


async def get_operation_logs(
    k8s_manager: AsyncK8sManager,
    k8s_operation: any,
    instance: str,
    last_time: Optional[datetime.datetime],
):
    previous_last = last_time
    operation_logs, last_time = await query_k8s_operation_logs(
        instance=instance,
        last_time=None,
        k8s_manager=k8s_manager,
        stream=True,
    )
    if k8s_operation["status"].get("completionTime"):
        last_time = None
    if previous_last:
        operation_logs = [
            l.to_dict() for l in operation_logs if l.timestamp > previous_last
        ]

    return operation_logs, last_time



try:
    from django.utils.crypto import salted_hmac
except ImportError:
    raise ImportError("This module depends on django.")


def get_hmac(key_salt, value):
    return salted_hmac(key_salt, value).hexdigest()



from django.contrib.auth.models import AbstractUser


class User(AbstractUser):
    class Meta(AbstractUser.Meta):
        app_label = "db"
        db_table = "db_user"
        swappable = "AUTH_USER_MODEL"



from django.db import models


class DescribableModel(models.Model):
    description = models.TextField(blank=True, null=True)

    class Meta:
        abstract = True

    @property
    def has_description(self) -> bool:
        return bool(self.description)



import os

from typing import Optional

from haupt.cli.runners.base import start_app
from polyaxon._services.values import PolyaxonServices


def start(
    host: Optional[str] = None,
    port: Optional[int] = None,
    log_level: Optional[str] = None,
    workers: Optional[int] = None,
    per_core: bool = False,
    uds: Optional[str] = None,
):
    start_app(
        app="haupt.polyconf.asgi.server:application",
        app_name=PolyaxonServices.API,
        host=host,
        port=port,
        log_level=log_level,
        workers=workers,
        per_core=per_core,
        uds=uds,
        migrate_tables=True,
        migrate_db=True,
    )



from rest_framework.relations import SlugRelatedField


class UuidSlugRelatedField(SlugRelatedField):
    def to_representation(self, obj):
        value = getattr(obj, self.slug_field)
        if value:
            return value.hex
        return value



from clipped.decorators.signals import ignore_raw, ignore_updates_pre

from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver

from haupt.db.defs import Models
from polyaxon._constants.metadata import META_IS_PROMOTED
from polyaxon.schemas import V1StageCondition, V1Stages


@receiver(
    pre_save, sender=Models.ProjectVersion, dispatch_uid="artifact_version_created"
)
@ignore_updates_pre
@ignore_raw
def artifact_version_created(sender, instance=None, created=False, **kwargs):
    instance.stage_conditions = [
        V1StageCondition.get_condition(
            type=V1Stages.TESTING,
            status="True",
            reason="{}VersionCreated".format(instance.kind.capitalize()),
            message="A new {} version is created".format(instance.kind),
        ).to_dict()
    ]


@receiver(
    pre_delete,
    sender=Models.ProjectVersion,
    dispatch_uid="check_related_run_to_artifact",
)
@ignore_raw
def check_related_run_to_artifact(sender, instance=None, created=False, **kwargs):
    if instance.run_id:
        run = instance.run
        if (
            META_IS_PROMOTED in run.meta_info
            and run.versions.exclude(id=instance.id).count() == 0
        ):
            run.meta_info.pop(META_IS_PROMOTED, None)
            run.save(update_fields=["meta_info"])



from haupt.common.options.option import Option, OptionScope, OptionStores

K8S_NAMESPACE = "K8S_NAMESPACE"
K8S_IN_CLUSTER = "K8S_IN_CLUSTER"

OPTIONS = {K8S_NAMESPACE, K8S_IN_CLUSTER}


class K8sNamespace(Option):
    key = K8S_NAMESPACE
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = False
    is_list = False
    typing = "str"
    store = OptionStores.SETTINGS
    default = None
    options = None


class K8sInCluster(Option):
    key = K8S_IN_CLUSTER
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = False
    is_list = False
    typing = "bool"
    store = OptionStores.SETTINGS
    default = None
    options = None



from rest_framework import status
from rest_framework.response import Response

from django.http import HttpRequest

from haupt.apis.serializers.runs import RunDetailSerializer
from haupt.common.apis.regex import OWNER_NAME_KEY, RUN_UUID_KEY, UUID_KEY
from haupt.common.endpoints.base import BaseEndpoint, PostEndpoint, RetrieveEndpoint
from haupt.db.defs import Models
from haupt.db.managers.agents import get_agent_state, trigger_cron
from polyaxon.schemas import LiveState, V1Statuses


class AgentStateViewV1(BaseEndpoint, RetrieveEndpoint):
    ALLOWED_METHODS = ["GET"]

    def get(self, request, *args, **kwargs):
        state = get_agent_state()
        return Response(
            data={
                "state": state,
                "status": V1Statuses.RUNNING,
                "live_state": LiveState.LIVE,
            },
            status=status.HTTP_200_OK,
        )


class AgentCronViewV1(BaseEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        trigger_cron()
        return Response(status=status.HTTP_200_OK)


class AgentRunDetailView(BaseEndpoint, RetrieveEndpoint):
    serializer_class = RunDetailSerializer
    queryset = Models.Run.all.select_related(
        "original",
        "pipeline",
        "project",
    ).prefetch_related("contributors")
    ALLOWED_METHODS = ["GET"]
    lookup_field = UUID_KEY
    lookup_url_kwarg = RUN_UUID_KEY
    CONTEXT_KEYS = (OWNER_NAME_KEY, RUN_UUID_KEY)
    CONTEXT_OBJECTS = ("run",)

    def initialize_object_context(self, request: HttpRequest, *args, **kwargs) -> None:
        #  pylint:disable=attribute-defined-outside-init
        self.run = self.get_object()



from rest_framework import serializers


class UserMixin(serializers.Serializer):
    def get_user(self, obj):
        return obj.user.username if obj.user_id else None


class UserEmailMixin(serializers.Serializer):
    def get_user_email(self, obj):
        return obj.user.email if obj.user_id else None



from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

from haupt.db.abstracts.diff import DiffModel


class StatsModel(DiffModel):
    created_at = models.DateTimeField(db_index=True)
    user = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )
    run = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )
    status = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )
    version = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )
    tracking_time = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )

    class Meta:
        abstract = True
        app_label = "db"



from typing import List, Optional

import click


@click.command()
@click.option(
    "-c",
    "--command",
    help="The command to execute.",
)
@click.argument("args", nargs=-1)
def manage(command: str, args: Optional[List[str]]):
    """Start a new sever session."""
    from haupt.cli.runners.manage import run_manage

    return run_manage(command, args)



from haupt import settings
from haupt.proxies.schemas.base import clean_config, get_config
from haupt.proxies.schemas.urls import (
    get_header_host,
    get_service_proxy,
    get_service_url,
    get_ssl_server_name,
)
from polyaxon.api import (
    ADMIN_V1_LOCATION,
    API_V1_LOCATION,
    AUTH_V1_LOCATION,
    SSO_V1_LOCATION,
    UI_V1_LOCATION,
)

OPTIONS = r"""
location {path} {{
    {auth}
    {resolver}
    {ssl_server_name}
    proxy_pass {service};
    proxy_http_version 1.1;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_set_header Origin "";
    proxy_set_header X-Real-IP $remote_addr;
    {header_host}
    proxy_buffering off;
}}
"""  # noqa


def get_api_config(
    path: str,
    service: str,
    resolver: str,
    auth: str,
    ssl_server_name: str,
    header_host: str,
):
    return get_config(
        options=OPTIONS,
        path=path,
        service=service,
        resolver=resolver,
        auth=auth,
        ssl_server_name=ssl_server_name,
        header_host=header_host,
    )


def get_api_location_config(resolver: str, auth=str):
    service = get_service_url(
        host=settings.PROXIES_CONFIG.api_host,
        port=settings.PROXIES_CONFIG.api_port,
    )
    if not settings.PROXIES_CONFIG.api_use_resolver:
        resolver = ""
    if not settings.PROXIES_CONFIG.auth_external:
        auth = ""
    ssl_server_name = get_ssl_server_name(service)
    header_host = get_header_host(service)
    if settings.PROXIES_CONFIG.has_forward_proxy:
        service = get_service_proxy()
    config = [
        get_api_config(
            path="= /",
            service=service,
            resolver=resolver,
            auth=auth,
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
        get_api_config(
            path=API_V1_LOCATION,
            service=service,
            resolver=resolver,
            auth=auth,
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
        get_api_config(
            path=AUTH_V1_LOCATION,
            service=service,
            resolver=resolver,
            auth="",
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
        get_api_config(
            path=UI_V1_LOCATION,
            service=service,
            resolver=resolver,
            auth="",
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
        get_api_config(
            path=SSO_V1_LOCATION,
            service=service,
            resolver=resolver,
            auth="",
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
        get_api_config(
            path="/static/",
            service=service,
            resolver=resolver,
            auth="",
            ssl_server_name=ssl_server_name,
            header_host=header_host,
        ),
    ]
    if settings.PROXIES_CONFIG.ui_admin_enabled:
        config.append(
            get_api_config(
                path=ADMIN_V1_LOCATION,
                service=service,
                resolver=resolver,
                auth=auth,
                ssl_server_name=ssl_server_name,
                header_host=header_host,
            )
        )
    return clean_config(config)



from haupt.common.events import event_actions, event_subjects
from haupt.common.events.event import ActorEvent, Attribute, Event
from haupt.common.events.registry.attributes import (
    PROJECT_RESOURCE_ATTRIBUTES,
    PROJECT_RESOURCE_OWNER_ATTRIBUTES,
    PROJECT_RUN_EXECUTOR_ATTRIBUTES,
    PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES,
)

RUN_CREATED = "{}.{}".format(event_subjects.RUN, event_actions.CREATED)
RUN_STOPPED = "{}.{}".format(event_subjects.RUN, event_actions.STOPPED)
RUN_SKIPPED = "{}.{}".format(event_subjects.RUN, event_actions.SKIPPED)
RUN_NEW_STATUS = "{}.{}".format(event_subjects.RUN, event_actions.NEW_STATUS)
RUN_NEW_ARTIFACTS = "{}.{}".format(event_subjects.RUN, event_actions.NEW_ARTIFACTS)
RUN_SUCCEEDED = "{}.{}".format(event_subjects.RUN, event_actions.SUCCEEDED)
RUN_FAILED = "{}.{}".format(event_subjects.RUN, event_actions.FAILED)
RUN_DONE = "{}.{}".format(event_subjects.RUN, event_actions.DONE)
RUN_RESUMED = "{}.{}".format(event_subjects.RUN, event_actions.RESUMED)
RUN_UPDATED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.UPDATED, event_subjects.ACTOR
)
RUN_CREATED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.CREATED, event_subjects.ACTOR
)
RUN_SYNCED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.SYNCED, event_subjects.ACTOR
)
RUN_VIEWED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.VIEWED, event_subjects.ACTOR
)
RUN_DELETED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.DELETED, event_subjects.ACTOR
)
RUN_STOPPED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.STOPPED, event_subjects.ACTOR
)
RUN_APPROVED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.APPROVED, event_subjects.ACTOR
)
RUN_TRANSFERRED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.TRANSFERRED, event_subjects.ACTOR
)
RUN_INVALIDATED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.INVALIDATED, event_subjects.ACTOR
)
RUN_RESUMED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.RESUMED, event_subjects.ACTOR
)
RUN_RESTARTED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.RESTARTED, event_subjects.ACTOR
)
RUN_COPIED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.COPIED, event_subjects.ACTOR
)
RUN_SKIPPED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.SKIPPED, event_subjects.ACTOR
)
RUN_STATS_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.STATS, event_subjects.ACTOR
)

EVENTS = {
    RUN_CREATED,
    RUN_STOPPED,
    RUN_RESUMED,
    RUN_SKIPPED,
    RUN_NEW_STATUS,
    RUN_NEW_ARTIFACTS,
    RUN_SUCCEEDED,
    RUN_FAILED,
    RUN_DONE,
    RUN_CREATED_ACTOR,
    RUN_SYNCED_ACTOR,
    RUN_UPDATED_ACTOR,
    RUN_VIEWED_ACTOR,
    RUN_DELETED_ACTOR,
    RUN_STOPPED_ACTOR,
    RUN_APPROVED_ACTOR,
    RUN_TRANSFERRED_ACTOR,
    RUN_INVALIDATED_ACTOR,
    RUN_RESUMED_ACTOR,
    RUN_RESTARTED_ACTOR,
    RUN_COPIED_ACTOR,
    RUN_SKIPPED_ACTOR,
    RUN_STATS_ACTOR,
}


class RunEvent(Event):
    entity_uuid = "project.uuid"


class RunActorEvent(ActorEvent):
    entity_uuid = "project.uuid"


class RunCreatedEvent(RunEvent):
    event_type = RUN_CREATED
    attributes = PROJECT_RUN_EXECUTOR_ATTRIBUTES


class RunStoppedEvent(Event):
    event_type = RUN_STOPPED
    attributes = PROJECT_RESOURCE_ATTRIBUTES


class RunResumedEvent(RunEvent):
    event_type = RUN_RESUMED
    attributes = PROJECT_RESOURCE_ATTRIBUTES


class RunSkippedEvent(RunEvent):
    event_type = RUN_SKIPPED
    attributes = PROJECT_RESOURCE_ATTRIBUTES


class RunNewStatusEvent(RunEvent):
    event_type = RUN_NEW_STATUS
    attributes = PROJECT_RESOURCE_ATTRIBUTES + (
        Attribute("status"),
        Attribute("previous_status", is_required=False),
    )


class RunNewArtifactsEvent(RunEvent):
    event_type = RUN_NEW_ARTIFACTS
    attributes = PROJECT_RESOURCE_ATTRIBUTES + (Attribute("artifacts", attr_type=list),)


class RunSucceededEvent(RunEvent):
    event_type = RUN_SUCCEEDED
    attributes = PROJECT_RESOURCE_ATTRIBUTES


class RunFailedEvent(RunEvent):
    event_type = RUN_FAILED
    attributes = PROJECT_RESOURCE_ATTRIBUTES


class RunDoneEvent(RunEvent):
    event_type = RUN_DONE
    attributes = PROJECT_RUN_EXECUTOR_ATTRIBUTES


class RunCreatedActorEvent(RunActorEvent):
    event_type = RUN_CREATED_ACTOR
    actor_id = "user.id"
    actor_name = "user.username"
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunSyncedActorEvent(RunActorEvent):
    event_type = RUN_SYNCED_ACTOR
    actor_id = "user.id"
    actor_name = "user.username"
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunUpdatedActorEvent(RunActorEvent):
    event_type = RUN_UPDATED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunDeletedActorEvent(RunActorEvent):
    event_type = RUN_DELETED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunViewedActorEvent(RunActorEvent):
    event_type = RUN_VIEWED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunStoppedActorEvent(RunActorEvent):
    event_type = RUN_STOPPED_ACTOR
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunApprovedActorEvent(RunActorEvent):
    event_type = RUN_APPROVED_ACTOR
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunTransferredActorEvent(RunActorEvent):
    event_type = RUN_TRANSFERRED_ACTOR
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunInvalidatedActorEvent(RunActorEvent):
    event_type = RUN_INVALIDATED_ACTOR
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunResumedActorEvent(RunActorEvent):
    event_type = RUN_RESUMED_ACTOR
    attributes = PROJECT_RUN_EXECUTOR_OWNER_ATTRIBUTES


class RunRestartedActorEvent(RunActorEvent):
    event_type = RUN_RESTARTED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunCopiedActorEvent(RunActorEvent):
    event_type = RUN_COPIED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunSkippedActorEvent(RunActorEvent):
    event_type = RUN_SKIPPED_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunStatsActorEvent(RunActorEvent):
    event_type = RUN_STATS_ACTOR
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES



"""
ASGI config for deploy project.

It exposes the ASGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/4.1/howto/deployment/asgi/
"""

import os

from django.core.asgi import get_asgi_application

from haupt import settings
from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.keys import ENV_KEYS_SERVICE, ENV_KEYS_UI_IN_SANDBOX
from polyaxon._services.values import PolyaxonServices

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "haupt.polyconf.settings")
os.environ.setdefault("ASGI_APPLICATION", "haupt.polyconf.asgi.sandbox.application")
os.environ[ENV_KEYS_SERVICE] = PolyaxonServices.API
os.environ[ENV_KEYS_UI_IN_SANDBOX] = "true"
settings.set_sandbox_config(path=ctx_paths.CONTEXT_ARTIFACTS_ROOT)
application = get_asgi_application()



from typing import Any

from haupt.common.authentication.base import BaseUser


class BotUser(BaseUser):
    NAME = "bot"

    def __init__(self, service: str, username: str):
        self.username = username
        self.is_bot = True
        self.is_sa = False
        super().__init__(service)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, BotUser) and other.username == self.username


def is_bot_user(user: Any) -> bool:
    return getattr(user, "is_bot", False)


def is_sa_user(user: Any) -> bool:
    return getattr(user, "is_sa", False)


def is_bot_or_sa_user(user: Any) -> bool:
    return is_bot_user(user) or is_sa_user(user)


def is_authenticated_bot_user(user: Any) -> bool:
    if is_bot_user(user):
        return user.is_bot

    return False


def is_authenticated_service_account(user: Any) -> bool:
    if is_sa_user(user):
        return user.is_sa

    return False



import factory

from haupt.db.defs import Models
from traceml.artifacts import V1ArtifactKind


class ArtifactFactory(factory.django.DjangoModelFactory):
    name = "accuracy"
    kind = V1ArtifactKind.METRIC
    summary = {"last_value": 0.9, "max_value": 0.9, "min_value": 0.1}
    path = "accuracy"

    class Meta:
        model = Models.Artifact



from haupt.background.celeryp.queues import CeleryCoreQueues
from haupt.background.celeryp.tasks import SchedulerCeleryTasks

SCHEDULER_CORE_ROUTES = {
    # Scheduler health
    SchedulerCeleryTasks.SCHEDULER_HEALTH: {"queue": CeleryCoreQueues.SCHEDULER_HEALTH},
    # compiler
    SchedulerCeleryTasks.RUNS_PREPARE: {"queue": CeleryCoreQueues.SCHEDULER_COMPILER},
    SchedulerCeleryTasks.RUNS_START: {"queue": CeleryCoreQueues.SCHEDULER_COMPILER},
    SchedulerCeleryTasks.RUNS_BUILT: {"queue": CeleryCoreQueues.SCHEDULER_COMPILER},
    # Scheduler runs
    SchedulerCeleryTasks.RUNS_STOP: {"queue": CeleryCoreQueues.SCHEDULER_RUNS},
    SchedulerCeleryTasks.RUNS_HOOKS: {"queue": CeleryCoreQueues.SCHEDULER_RUNS},
    # Scheduler artifacts
    SchedulerCeleryTasks.RUNS_SET_ARTIFACTS: {
        "queue": CeleryCoreQueues.SCHEDULER_ARTIFACTS,
        "priority": 0,
    },
}


def get_routes():
    routes = {}
    routes.update(SCHEDULER_CORE_ROUTES)
    return routes



from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from polyaxon.schemas import V1RunKind


async def get_op_spec(
    k8s_manager: AsyncK8sManager,
    run_uuid: str,
    run_kind: str,
):
    pods = await k8s_manager.list_pods(
        label_selector=k8s_manager.get_managed_by_polyaxon(run_uuid)
    )
    pods_list = {}
    for pod in pods or []:
        pods_list[
            pod.metadata.name
        ] = k8s_manager.api_client.sanitize_for_serialization(pod)
    data = {"pods": pods_list}
    if V1RunKind.has_service(run_kind):
        services = await k8s_manager.list_services(
            label_selector=k8s_manager.get_managed_by_polyaxon(run_uuid)
        )
        services_list = {}
        for service in services or []:
            services_list[
                service.metadata.name
            ] = k8s_manager.api_client.sanitize_for_serialization(service)
        data["services"] = services_list
    return data



from datetime import datetime
from typing import List

from clipped.utils.json import orjson_loads
from clipped.utils.paths import delete_path

from asgiref.sync import sync_to_async
from polyaxon._fs.async_manager import (
    delete_file_or_dir,
    download_dir,
    open_file,
    upload_data,
)
from polyaxon._fs.types import FSSystem
from traceml.logging import V1Log, V1Logs


async def clean_tmp_logs(fs: FSSystem, store_path: str, run_uuid: str):
    subpath = "{}/.tmpplxlogs".format(run_uuid)
    delete_path(subpath)
    await delete_file_or_dir(
        fs=fs, store_path=store_path, subpath=subpath, is_file=False
    )


async def upload_logs(fs: FSSystem, store_path: str, run_uuid: str, logs: List[V1Log]):
    for c_logs in V1Logs.chunk_logs(logs):
        last_file = datetime.timestamp(c_logs.logs[-1].timestamp)
        subpath = "{}/plxlogs/{}".format(run_uuid, last_file)
        await upload_data(
            fs=fs, store_path=store_path, subpath=subpath, data=c_logs.to_json()
        )


async def content_to_logs(content, logs_path):
    if not content:
        return []

    @sync_to_async
    def convert():
        # Version handling
        if ".plx" in logs_path:
            return V1Logs.read_csv(content).logs
        # Chunked logs
        return orjson_loads(content).get("logs", [])

    return await convert()


async def download_logs_file(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    last_file: str,
    check_cache: bool = True,
) -> (str, str):
    subpath = "{}/plxlogs/{}".format(run_uuid, last_file)
    content = await open_file(
        fs=fs, store_path=store_path, subpath=subpath, check_cache=check_cache
    )

    return await content_to_logs(content, subpath)


async def download_tmp_logs(fs: FSSystem, store_path: str, run_uuid: str) -> str:
    subpath = "{}/.tmpplxlogs".format(run_uuid)
    delete_path(subpath)
    return await download_dir(fs=fs, store_path=store_path, subpath=subpath)



import logging
import os

from typing import Dict, List, Optional, Set

from clipped.utils.json import orjson_loads

import aiofiles

from asgiref.sync import sync_to_async
from polyaxon._fs.async_manager import (
    download_file,
    download_files,
    list_files,
    tar_files,
)
from polyaxon._fs.types import FSSystem
from traceml.artifacts import V1ArtifactKind
from traceml.events import V1Events, get_event_path, get_resource_path

logger = logging.getLogger("haupt.streams.events")


async def get_events_files(
    fs: FSSystem, store_path: str, run_uuid: str, event_kind: str
) -> List[str]:
    subpath = get_event_path(run_path=run_uuid, kind=event_kind)
    files = await list_files(fs=fs, store_path=store_path, subpath=subpath)
    if not files["files"]:
        return []
    return sorted([f for f in files["files"].keys()])


async def get_resources_files(
    fs: FSSystem, store_path: str, run_uuid: str
) -> List[str]:
    subpath = get_resource_path(run_path=run_uuid, kind=V1ArtifactKind.METRIC)
    files = await list_files(fs=fs, store_path=store_path, subpath=subpath)
    if not files["files"]:
        return []
    return sorted([f for f in files["files"].keys()])


async def process_operation_event(
    event_path: str,
    event_kind: str,
    event_name: str,
    orient: str = V1Events.ORIENT_CSV,
    sample: Optional[int] = None,
    to_dict: bool = True,
) -> Optional[Dict]:
    if not event_path or not os.path.exists(event_path):
        return None

    async with aiofiles.open(event_path, mode="r") as f:
        contents = await f.read()
        if contents:
            if orient == V1Events.ORIENT_CSV:
                return {"name": event_name, "kind": event_kind, "data": contents}
            if orient == V1Events.ORIENT_DICT:
                event_df = await sync_to_async(V1Events.read)(
                    kind=event_kind,
                    name=event_name,
                    data=contents,
                    parse_dates=False,
                    engine="pyarrow",
                )
                if sample:
                    try:
                        sample = int(sample)
                        if event_df.df.shape[0] > sample:
                            event_df.df = event_df.df.sample(
                                n=sample, random_state=0
                            ).sort_index()
                    except Exception as e:
                        logger.warning("Could not sample event dataframe, error %s", e)
                return {
                    "name": event_name,
                    "kind": event_kind,
                    "data": event_df.to_dict() if to_dict else event_df,
                }
            else:
                logger.warning(
                    "received an unrecognisable orient value {}.".format(orient)
                )
    return None


async def get_archived_operation_resource(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_name: str,
    orient: str = V1Events.ORIENT_CSV,
    check_cache: bool = True,
    sample: Optional[int] = None,
) -> Optional[Dict]:
    subpath = get_resource_path(run_path=run_uuid, kind=event_kind, name=event_name)
    event_path = await download_file(
        fs=fs, store_path=store_path, subpath=subpath, check_cache=check_cache
    )

    return await process_operation_event(
        event_path=event_path,
        event_kind=event_kind,
        event_name=event_name,
        orient=orient,
        sample=sample,
    )


async def get_archived_operation_event(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_name: str,
    orient: str = V1Events.ORIENT_CSV,
    check_cache: bool = True,
    sample: Optional[int] = None,
) -> Optional[Dict]:
    subpath = get_event_path(run_path=run_uuid, kind=event_kind, name=event_name)
    event_path = await download_file(
        fs=fs, store_path=store_path, subpath=subpath, check_cache=check_cache
    )

    return await process_operation_event(
        event_path=event_path,
        event_kind=event_kind,
        event_name=event_name,
        orient=orient,
        sample=sample,
    )


async def get_archived_operation_event_and_assets(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_name: str,
    check_cache: bool = True,
) -> List[str]:
    pkg_files = []
    subpath = get_event_path(run_path=run_uuid, kind=event_kind, name=event_name)
    event_path = await download_file(
        fs=fs, store_path=store_path, subpath=subpath, check_cache=check_cache
    )
    pkg_files.append(event_path)

    event = await process_operation_event(
        event_path=event_path,
        event_kind=event_kind,
        event_name=event_name,
        orient=V1Events.ORIENT_DICT,
        sample=None,
        to_dict=False,
    )
    if not event:
        logger.warning("During the packaging of %s, the event download failed.")
        return []
    df = event["data"].df
    try:
        files = df[event_kind].map(lambda f: orjson_loads(f).get("path")).tolist()
    except Exception as e:
        logger.warning(
            "During the packaging of %s, the event format found was not correct. "
            "Error %s" % (event_path, e)
        )
        return pkg_files
    subpaths = []
    for file_from_path in files:
        subpaths.append("{}/{}".format(run_uuid, file_from_path))

    return await download_files(
        fs=fs, subpaths=subpaths, check_cache=check_cache, pkg_files=pkg_files
    )


async def get_archived_operation_events_and_assets(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_names: Set[str],
    check_cache: bool = True,
) -> Optional[str]:
    pkg_files = []
    for event_name in event_names:
        event_pkg_files = await get_archived_operation_event_and_assets(
            fs=fs,
            store_path=store_path,
            run_uuid=run_uuid,
            event_kind=event_kind,
            event_name=event_name,
            check_cache=check_cache,
        )
        pkg_files += event_pkg_files
    return await tar_files(
        filename="{}.{}.{}".format(run_uuid, event_kind, "-and-".join(event_names)),
        pkg_files=pkg_files,
        subpath=run_uuid,
    )


async def get_archived_operation_resources(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_names: Set[str],
    orient: str = V1Events.ORIENT_CSV,
    check_cache: bool = True,
    sample: Optional[int] = None,
) -> List[Dict]:
    events = []
    if not event_names:
        files = await get_resources_files(
            fs=fs, store_path=store_path, run_uuid=run_uuid
        )
        event_names = [f.split(".plx")[0] for f in files]
    for event_name in event_names:
        event = await get_archived_operation_resource(
            fs=fs,
            store_path=store_path,
            run_uuid=run_uuid,
            event_kind=event_kind,
            event_name=event_name,
            orient=orient,
            check_cache=check_cache,
            sample=sample,
        )
        if event:
            events.append(event)
    return events


async def get_archived_operation_events(
    fs: FSSystem,
    store_path: str,
    run_uuid: str,
    event_kind: str,
    event_names: Set[str],
    orient: str = V1Events.ORIENT_CSV,
    check_cache: bool = True,
    sample: Optional[int] = None,
) -> List[Dict]:
    events = []
    for event_name in event_names:
        event = await get_archived_operation_event(
            fs=fs,
            store_path=store_path,
            run_uuid=run_uuid,
            event_kind=event_kind,
            event_name=event_name,
            orient=orient,
            check_cache=check_cache,
            sample=sample,
        )
        if event:
            events.append(event)
    return events


async def get_archived_operations_events(
    fs: FSSystem,
    store_path: str,
    event_kind: str,
    run_uuids: Set[str],
    event_names: Set[str],
    orient: str = V1Events.ORIENT_CSV,
    check_cache: bool = True,
    sample: Optional[int] = None,
) -> Dict[str, List]:
    events = {}
    for run_uuid in run_uuids:
        run_events = await get_archived_operation_events(
            fs=fs,
            store_path=store_path,
            run_uuid=run_uuid,
            event_kind=event_kind,
            event_names=event_names,
            orient=orient,
            check_cache=check_cache,
            sample=sample,
        )
        events[run_uuid] = run_events
    return events



import factory

from haupt.db.defs import Models
from haupt.db.factories.projects import ProjectFactory
from haupt.db.factories.users import UserFactory
from polyaxon.schemas import ManagedBy


class RunFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    project = factory.SubFactory(ProjectFactory)
    original = None
    pipeline = None
    managed_by = ManagedBy.USER

    class Meta:
        model = Models.Run



import django.core.serializers.json
import django.db.models.deletion

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0013_major_upgrade"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="run",
            name="is_managed",
        ),
        migrations.AddField(
            model_name="project",
            name="archived_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="project",
            name="contributors",
            field=models.ManyToManyField(
                blank=True, related_name="+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="project",
            name="deleted_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="projectversion",
            name="contributors",
            field=models.ManyToManyField(
                blank=True, related_name="+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="archived_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="run",
            name="contributors",
            field=models.ManyToManyField(
                blank=True, related_name="+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="deleted_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name="ProjectStats",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                ("created_at", models.DateTimeField(db_index=True)),
                (
                    "user",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "run",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "tracking_time",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stats",
                        to="db.project",
                    ),
                ),
            ],
            options={
                "db_table": "db_projectstats",
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="project",
            name="latest_stats",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="db.projectstats",
            ),
        ),
    ]



import copy

from clipped.utils.imports import import_string

from haupt.common.auditor.manager import event_manager
from haupt.common.events.event import Event
from haupt.common.events.event_service import EventService


class AuditorService(EventService):
    """A service that just passes the event to other services."""

    event_manager = event_manager

    def __init__(
        self, auditor_events_task=None, workers_service=None, executor_service=None
    ):
        self.auditor_events_task = auditor_events_task
        self.executor = None
        self.workers = None
        self.workers_service = workers_service
        self.executor_service = executor_service

    def record_event(self, event: Event):
        """
        Record the event async.
        """
        serialized_event = event.serialize(
            dumps=False, include_actor_name=True, include_instance_info=True
        )
        if self.workers and self.auditor_events_task:
            self.workers.send(
                self.auditor_events_task,
                kwargs={"event": copy.deepcopy(serialized_event)},
            )

        if self.executor:
            # We include the instance in the serialized event for executor
            serialized_event["instance"] = event.instance
            self.executor.record(
                event_type=event.event_type, event_data=serialized_event
            )

    def setup(self):
        super().setup()

        if self.workers_service:
            self.workers = import_string(self.workers_service)
        if self.executor_service:
            self.executor = import_string(self.executor_service)



import os

from haupt.schemas.sandbox_config import SandboxConfig
from polyaxon._config.manager import ConfigManager


class SandboxConfigManager(ConfigManager):
    """Manages sandbox configuration .sandbox file."""

    VISIBILITY = ConfigManager.Visibility.ALL
    CONFIG_FILE_NAME = ".sandbox"
    CONFIG = SandboxConfig

    @classmethod
    def get_config_from_env(cls):
        config_paths = [os.environ, {"dummy": "dummy"}]

        config = cls._CONFIG_READER.read_configs(config_paths)
        config = cls.CONFIG.from_dict(config.data)
        config.set_default_artifacts_store()
        return config

    @classmethod
    def get_config_or_default(cls) -> SandboxConfig:
        if not cls.is_initialized():
            return cls.get_config_from_env()

        return cls.get_config()



from haupt.common.settings.apps import set_apps
from haupt.common.settings.assets import set_assets
from haupt.common.settings.core import set_core
from haupt.common.settings.cors import set_cors
from haupt.common.settings.middlewares import set_base_middlewares
from haupt.common.settings.ui import set_ui
from haupt.schemas.platform_config import PlatformConfig


def set_streams_apps(context, config: PlatformConfig):
    set_apps(
        context=context,
        config=config,
        third_party_apps=("rest_framework", "corsheaders"),
        project_apps=(
            "haupt.common.apis.apps.CommonApisConfig",
            "haupt.streams.apps.StreamsConfig",
        ),
        use_db_apps=False,
        use_staticfiles_app=context["UI_IN_SANDBOX"],
    )


def set_service(context, config: PlatformConfig):
    # This is repeated because it's required for using the staticfiles app
    context["UI_IN_SANDBOX"] = config.ui_in_sandbox
    set_streams_apps(context, config)
    set_core(context=context, config=config, use_db=False)
    set_cors(context=context, config=config)
    set_ui(context=context, config=config)
    set_base_middlewares(context=context, config=config)
    set_assets(context=context, config=config)



from typing import Dict, List, Optional, Set

from django.conf import settings
from django.db.models import Count, Q

from haupt.common.authentication.base import is_normal_user
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from polyaxon.schemas import (
    LifeCycle,
    ManagedBy,
    V1CompiledOperation,
    V1RunKind,
    V1RunPending,
    V1StatusCondition,
    V1Statuses,
)


def add_run_contributors(
    run: Models.Run,
    users: Optional[List[Models.User]] = None,
    user_ids: Optional[List[int]] = None,
):
    if not settings.HAS_ORG_MANAGEMENT:
        return
    if not run:
        return
    _users = [u.id for u in users if is_normal_user(u)] if users else user_ids
    if not _users:
        return

    run.contributors.add(*_users)


def create_run(
    project_id: int,
    user_id: int,
    name: Optional[str] = None,
    description: Optional[str] = None,
    readme: Optional[str] = None,
    tags: List[int] = None,
    raw_content: Optional[str] = None,
    meta_info: Optional[Dict] = None,
    managed_by: Optional[ManagedBy] = ManagedBy.USER,
) -> BaseRun:
    instance = Models.Run.objects.create(
        project_id=project_id,
        user_id=user_id,
        name=name,
        description=description,
        readme=readme,
        tags=tags,
        kind=V1RunKind.JOB,
        managed_by=managed_by,
        raw_content=raw_content,
        meta_info=meta_info,
        status_conditions=[
            V1StatusCondition.get_condition(
                type=V1Statuses.CREATED,
                status="True",
                reason="ModelManager",
                message="Run is created",
            ).to_dict()
        ],
    )
    return instance


def base_approve_run(run: BaseRun):
    pending = run.pending
    if pending:
        new_pending = None
        if (
            (pending == V1RunPending.BUILD and run.status == V1Statuses.CREATED)
            or pending == V1RunPending.UPLOAD
        ) and run.content:
            compiled_operation = V1CompiledOperation.read(
                run.content
            )  # TODO: Use construct
            if compiled_operation.is_approved is False:
                new_pending = V1RunPending.APPROVAL
        run.pending = new_pending
        run.save(update_fields=["pending", "updated_at"])


def get_runs_for(
    run_id: int,
    status: Optional[str],
    statuses: Optional[Set[str]],
    controller: bool = False,
    distinct: bool = True,
):
    filters = {}
    if controller:
        filters["controller_id"] = run_id
    else:
        filters["pipeline_id"] = run_id
    if status:
        filters["status"] = status
    if statuses:
        filters["status__in"] = statuses
    query = Models.Run.objects.filter(**filters)
    if distinct:
        query = query.distinct()

    return query


def is_pipeline_done(run: BaseRun) -> bool:
    return not run.pipeline_runs.exclude(status__in=LifeCycle.DONE_VALUES).exists()


def get_scheduled_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id, status=V1Statuses.SCHEDULED, statuses=None, controller=controller
    )


def get_succeeded_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id, status=V1Statuses.SUCCEEDED, statuses=None, controller=controller
    )


def get_failed_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id, status=V1Statuses.FAILED, statuses=None, controller=controller
    )


def get_stopped_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id, status=V1Statuses.STOPPED, statuses=None, controller=controller
    )


def get_done_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id,
        status=None,
        statuses=LifeCycle.DONE_VALUES,
        controller=controller,
    )


def get_pending_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id,
        status=None,
        statuses=LifeCycle.PENDING_VALUES,
        controller=controller,
    )


def get_running_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id,
        status=None,
        statuses=LifeCycle.RUNNING_VALUES,
        controller=controller,
    )


def get_on_k8s_runs(run_id: int, controller: bool = False):
    return get_runs_for(
        run_id=run_id,
        status=None,
        statuses=LifeCycle.ON_K8S_VALUES,
        controller=controller,
    )


def get_failed_stopped_and_all_runs(run_id: int, controller: bool = False):
    query = get_runs_for(
        run_id=run_id,
        status=None,
        statuses=None,
        controller=controller,
        distinct=False,
    )
    return query.aggregate(
        all=Count(
            "id",
            distinct=True,
        ),
        failed=Count(
            "id",
            filter=Q(status=V1Statuses.FAILED),
            distinct=True,
        ),
        stopped=Count(
            "id",
            filter=Q(status=V1Statuses.STOPPED),
            distinct=True,
        ),
    )


def get_stopping_pipelines_with_no_runs(queryset):
    return (
        queryset.filter(
            kind__in=[V1RunKind.DAG, V1RunKind.MATRIX, V1RunKind.SCHEDULE],
            status=V1Statuses.STOPPING,
        )
        .annotate(
            unfinished=Count(
                "pipeline_runs",
                filter=~Q(
                    pipeline_runs__status__in=LifeCycle.DONE_VALUES
                    | LifeCycle.PENDING_VALUES
                ),
                distinct=True,
            )
        )
        .filter(unfinished=0)
    )



from typing import Optional, Tuple

from haupt.schemas.platform_config import PlatformConfig


def set_apps(
    context,
    config: PlatformConfig,
    third_party_apps: Optional[Tuple],
    project_apps: Tuple,
    db_app: Optional[str] = None,
    use_db_apps: bool = True,
    use_admin_apps: bool = False,
    use_staticfiles_app: bool = True,
):
    extra_apps = tuple(config.extra_apps) if config.extra_apps else ()

    apps = ()
    if use_db_apps:
        apps += (
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
        )
        if use_admin_apps:
            apps += (
                "django.contrib.admin",
                "django.contrib.admindocs",
            )

    if use_staticfiles_app:
        apps += ("django.contrib.staticfiles",)

    model_apps = ()
    if use_db_apps:
        if db_app:
            model_apps = (db_app,)
        elif config.is_sqlite_db_engine:
            model_apps = ("haupt.db.sqlite.db.apps.DBConfig",)
        else:
            model_apps = ("haupt.db.pgsql.db.apps.DBConfig",)

    third_party_apps = third_party_apps or ()
    project_apps = project_apps or ()

    context["INSTALLED_APPS"] = (
        apps + third_party_apps + model_apps + extra_apps + project_apps
    )



CREATED = "created"
UPDATED = "updated"
DELETED = "deleted"
PROMOTED = "promoted"
VIEWED = "viewed"
STATS = "stats"
CLONED = "cloned"
RESUMED = "resumed"
STARTED = "started"
RESTARTED = "restarted"
COPIED = "copied"
SUCCEEDED = "succeeded"
FAILED = "failed"
SYNCED = "synced"
DONE = "done"
STOPPED = "stopped"
APPROVED = "approved"
TRANSFERRED = "transferred"
INVALIDATED = "invalidated"
SKIPPED = "skipped"
SETTINGS = "settings"
BILLING = "billing"
SUPPORT_REQUESTED = "support_requested"
NEW_STATUS = "new_status"
NEW_STAGE = "new_stage"
NEW_ARTIFACTS = "new_artifacts"
ARCHIVED = "archived"
RESTORED = "restored"
GRANTED = "granted"
REVOKED = "revoked"
INVITED = "invited"
DENIED = "denied"
REGISTERED = "registered"
ACTIVATED = "activated"
DOWNLOADED = "downloaded"
BOOKMARKED = "bookmarked"
UNBOOKMARKED = "unbookmarked"
ARTIFACTS_DOWNLOADED = "artifacts_downloaded"
RESOURCES_VIEWED = "resources_viewed"
LOGS_VIEWED = "logs_viewed"
STATUSES_VIEWED = "statuses_viewed"
RUNS_VIEWED = "runs_viewed"
METRICS_VIEWED = "metrics_viewed"
PROJECTS_VIEWED = "projects_viewed"
EXECUTED = "executed"
AUTH = "auth"
LDAP = "ldap"
GITHUB = "github"
GITLAB = "gitlab"
BITBUCKET = "bitbucket"
GOOGLE = "google"


WRITE_ACTIONS = [
    CREATED,
    UPDATED,
    DELETED,
    RESUMED,
    RESTARTED,
    SYNCED,
    COPIED,
    CLONED,
    STOPPED,
    APPROVED,
    TRANSFERRED,
    SETTINGS,
    PROMOTED,
    INVALIDATED,
    BILLING,
    SUPPORT_REQUESTED,
    ARCHIVED,
    RESTORED,
    GRANTED,
    REVOKED,
    INVITED,
]



"""
ASGI config for deploy project.

It exposes the ASGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/4.1/howto/deployment/asgi/
"""

import os

from django.core.asgi import get_asgi_application

from haupt import settings
from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.keys import (
    ENV_KEYS_SANDBOX_IS_LOCAL,
    ENV_KEYS_SERVICE,
    ENV_KEYS_UI_IN_SANDBOX,
)
from polyaxon._services.values import PolyaxonServices

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "haupt.polyconf.settings")
os.environ.setdefault("ASGI_APPLICATION", "haupt.polyconf.asgi.viewer.application")
os.environ[ENV_KEYS_SERVICE] = PolyaxonServices.STREAMS
os.environ[ENV_KEYS_UI_IN_SANDBOX] = "true"
os.environ[ENV_KEYS_SANDBOX_IS_LOCAL] = "true"
settings.set_sandbox_config(path=ctx_paths.CONTEXT_OFFLINE_ROOT)
application = get_asgi_application()



from typing import Optional

from clipped.compact.pydantic import Field, StrictInt, StrictStr, root_validator
from clipped.utils.http import clean_host

from polyaxon._env_vars.keys import (
    ENV_KEYS_K8S_NAMESPACE,
    ENV_KEYS_SANDBOX_DEBUG,
    ENV_KEYS_SANDBOX_HOST,
    ENV_KEYS_SANDBOX_IS_LOCAL,
    ENV_KEYS_SANDBOX_PER_CORE,
    ENV_KEYS_SANDBOX_PORT,
    ENV_KEYS_SANDBOX_SSL_ENABLED,
    ENV_KEYS_SANDBOX_WORKERS,
)
from polyaxon._schemas.agent import AgentConfig


class SandboxConfig(AgentConfig):
    _IDENTIFIER = "sandbox"

    namespace: Optional[StrictStr] = Field(
        alias=ENV_KEYS_K8S_NAMESPACE, default="sandbox"
    )
    port: Optional[StrictInt] = Field(alias=ENV_KEYS_SANDBOX_PORT)
    host: Optional[StrictStr] = Field(alias=ENV_KEYS_SANDBOX_HOST)
    ssl_enabled: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_SSL_ENABLED)
    debug: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_DEBUG)
    workers: Optional[StrictInt] = Field(alias=ENV_KEYS_SANDBOX_WORKERS)
    per_core: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_PER_CORE)
    is_local: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_IS_LOCAL)

    def __init__(
        self,
        host: Optional[str] = None,
        **data,
    ):
        super().__init__(
            host=clean_host(host) if host else host,
            **data,
        )

    @root_validator(pre=True)
    def handle_camel_case_sandbox(cls, values):
        if (
            not values.get("ssl_enabled")
            and not values.get(ENV_KEYS_SANDBOX_SSL_ENABLED)
            and "sslEnabled" in values
        ):
            values[ENV_KEYS_SANDBOX_SSL_ENABLED] = values["sslEnabled"]
        if (
            not values.get("per_core")
            and not values.get(ENV_KEYS_SANDBOX_PER_CORE)
            and "perCore" in values
        ):
            values[ENV_KEYS_SANDBOX_PER_CORE] = values["perCore"]
        if (
            not values.get("is_local")
            and not values.get(ENV_KEYS_SANDBOX_IS_LOCAL)
            and "isLocal" in values
        ):
            values[ENV_KEYS_SANDBOX_IS_LOCAL] = values["isLocal"]
        return values



from typing import Dict, Optional, Union

from django.core.handlers.asgi import ASGIRequest
from django.db import transaction
from django.http import HttpResponse
from django.urls import path

from haupt.common.endpoints.files import FilePathResponse
from haupt.common.endpoints.validation import validate_methods
from haupt.streams.connections.fs import AppFS
from haupt.streams.controllers.k8s_crd import get_k8s_operation
from haupt.streams.controllers.k8s_op_spec import get_op_spec
from haupt.streams.endpoints.base import UJSONResponse
from haupt.streams.tasks.op_spec import download_op_spec
from polyaxon import settings
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from polyaxon._utils.fqn_utils import get_resource_name_for_kind
from polyaxon.schemas import LifeCycle


@transaction.non_atomic_requests
async def k8s_inspect(
    request: ASGIRequest,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    methods: Optional[Dict] = None,
) -> Union[HttpResponse, FilePathResponse]:
    validate_methods(request, methods)
    resource_name = get_resource_name_for_kind(run_uuid=run_uuid)
    connection = request.GET.get("connection")
    status = request.GET.get("status")
    if LifeCycle.is_done(status):
        spec_path = await download_op_spec(
            fs=await AppFS.get_fs(connection=connection),
            store_path=AppFS.get_fs_root_path(connection=connection),
            run_uuid=run_uuid,
        )
        if spec_path:
            return FilePathResponse(filepath=spec_path)

    k8s_manager = AsyncK8sManager(
        namespace=namespace,
        in_cluster=settings.CLIENT_CONFIG.in_cluster,
    )
    await k8s_manager.setup()
    k8s_operation = await get_k8s_operation(
        k8s_manager=k8s_manager, resource_name=resource_name
    )
    data = None
    if k8s_operation:
        run_kind = k8s_operation["metadata"]["annotations"][
            "operation.polyaxon.com/kind"
        ]
        data = await get_op_spec(
            k8s_manager=k8s_manager, run_uuid=run_uuid, run_kind=run_kind
        )
    if k8s_manager:
        await k8s_manager.close()
    return UJSONResponse(data or {})


URLS_RUNS_K8S_INSPECT = (
    "<str:namespace>/<str:owner>/<str:project>/runs/<str:run_uuid>/k8s_inspect"
)

# fmt: off
k8s_routes = [
    path(
        URLS_RUNS_K8S_INSPECT,
        k8s_inspect,
        name="k8s_inspect",
        kwargs=dict(methods=["GET"]),
    ),
]



import os

from typing import Optional

from haupt import settings
from haupt.cli.runners.base import start_app
from polyaxon import settings as plx_settings
from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.keys import ENV_KEYS_SERVICE
from polyaxon._services.values import PolyaxonServices


def start(
    host: Optional[str] = None,
    port: Optional[int] = None,
    log_level: Optional[str] = None,
    workers: Optional[int] = None,
    per_core: bool = False,
    uds: Optional[str] = None,
):
    os.environ[ENV_KEYS_SERVICE] = PolyaxonServices.API
    settings.set_sandbox_config(path=ctx_paths.CONTEXT_ARTIFACTS_ROOT)
    host = host or settings.SANDBOX_CONFIG.host
    port = port or settings.SANDBOX_CONFIG.port
    start_app(
        app="haupt.polyconf.asgi.sandbox:application",
        app_name=PolyaxonServices.SANDBOX,
        host=host,
        port=port,
        log_level=log_level or plx_settings.CLIENT_CONFIG.log_level,
        workers=workers or settings.SANDBOX_CONFIG.workers,
        per_core=per_core or settings.SANDBOX_CONFIG.per_core,
        uds=uds,
        migrate_tables=True,
        migrate_db=True,
        use_cron=True,
    )



NAME = "haupt"
VERSION = "2.0.0-rc48"
DESC = (
    "Lineage metadata API, artifacts streams, sandbox, ML-API, and spaces for Polyaxon."
)
URL = "https://github.com/polyaxon/haupt"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "AGPLv3"



from haupt.db.abstracts.artifacts import BaseArtifact, BaseArtifactLineage


class Artifact(BaseArtifact):
    class Meta:
        app_label = "db"
        unique_together = (("name", "state"),)
        db_table = "db_artifact"


class ArtifactLineage(BaseArtifactLineage):
    class Meta:
        app_label = "db"
        unique_together = (("run", "artifact", "is_input"),)
        db_table = "db_artifactlineage"



# isort: skip_file

from django.apps import AppConfig


class APIsConfig(AppConfig):
    name = "haupt.apis"
    verbose_name = "APIs"

    def ready(self):
        from haupt.common import conf
        from haupt.common import auditor
        from haupt.orchestration import executor, operations
        from haupt.common import query

        conf.validate_and_setup()
        query.validate_and_setup()
        operations.validate_and_setup()
        executor.validate_and_setup()
        auditor.validate_and_setup()

        import haupt.db.signals.runs  # noqa
        import haupt.db.signals.versions  # noqa
        import haupt.db.signals.bookmarks  # noqa
        import haupt.db.signals.stats  # noqa

        import haupt.common.options.conf_subscriptions  # noqa
        from haupt.common.events import auditor_subscriptions  # noqa
        from haupt.db.administration import register  # noqa



import uuid

from datetime import datetime, timedelta
from functools import reduce
from operator import or_
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from clipped.config.patch_strategy import PatchStrategy
from clipped.utils.bools import to_bool
from clipped.utils.lists import to_list

from django.conf import settings
from django.db.models import Count, Q
from django.utils.timezone import now

from croniter import croniter
from haupt.common.db.runs import bulk_create_runs
from haupt.common.exceptions import AccessNotAuthorized, AccessNotFound
from haupt.db.abstracts.project_versions import BaseProjectVersion
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from haupt.db.managers.artifacts import set_artifacts
from haupt.db.managers.cache import get_run_state
from haupt.db.managers.statuses import new_run_status
from haupt.db.managers.versions import get_component_version_state
from haupt.db.query_managers.run import RunsOfflineFilter
from haupt.orchestration import operations
from hypertune.search_managers.grid_search.manager import GridSearchManager
from hypertune.search_managers.mapping.manager import MappingManager
from hypertune.search_managers.random_search.manager import RandomSearchManager
from polyaxon._auxiliaries import V1PolyaxonSidecarContainer
from polyaxon._compiler import resolver
from polyaxon._compiler.lineage import collect_lineage_artifacts_path
from polyaxon._constants.metadata import (
    META_CONCURRENCY,
    META_COPY_ARTIFACTS,
    META_DESTINATION_IMAGE,
    META_EDGE,
    META_HAS_DAGS,
    META_HAS_DOWNSTREAM_EVENTS_TRIGGER,
    META_HAS_EARLY_STOPPING,
    META_HAS_JOBS,
    META_HAS_MATRICES,
    META_HAS_SCHEDULES,
    META_HAS_SERVICES,
    META_IS_EXTERNAL,
    META_PORTS,
    META_REWRITE_PATH,
    META_UPLOAD_ARTIFACTS,
)
from polyaxon._contexts import keys as ctx_keys
from polyaxon._contexts import paths as ctx_paths
from polyaxon._contexts import refs as ctx_refs
from polyaxon._contexts import sections as ctx_sections
from polyaxon._env_vars.getters import get_versioned_entity_info
from polyaxon._polyaxonfile import CompiledOperationSpecification
from polyaxon._polyaxonfile.manager import (
    get_op_from_schedule,
    get_ops_from_suggestions,
)
from polyaxon._utils.fqn_utils import (
    get_project_instance,
    get_run_instance,
    get_run_lineage_paths,
)
from polyaxon.exceptions import PolyaxonCompilerError, PQLException
from polyaxon.schemas import (
    DagOpSpec,
    LifeCycle,
    V1ArtifactsType,
    V1CloningKind,
    V1CompiledOperation,
    V1Init,
    V1Join,
    V1JoinParam,
    V1Operation,
    V1OptimizationMetric,
    V1Param,
    V1Plugins,
    V1ProjectVersionKind,
    V1RunEdgeKind,
    V1RunPending,
    V1StatusCondition,
    V1Statuses,
    ops_params,
)


class SchedulingResolver(resolver.BaseResolver):
    def __init__(
        self,
        run: BaseRun,
        compiled_operation: V1CompiledOperation,
        owner_name: str,
        project_name: str,
        project_uuid: str,
        run_name: str,
        run_uuid: str,
        run_path: str,
        params: Optional[Dict],
        created_at: Optional[datetime] = None,
        compiled_at: Optional[datetime] = None,
        cloning_kind: V1CloningKind = None,
        original_uuid: Optional[str] = None,
        is_independent: bool = True,
    ):
        super().__init__(
            run=run,
            compiled_operation=compiled_operation,
            owner_name=owner_name,
            project_name=project_name,
            project_uuid=project_uuid,
            run_name=run_name,
            run_uuid=run_uuid,
            run_path=run_path,
            params=params,
            created_at=created_at,
            compiled_at=compiled_at,
            cloning_kind=cloning_kind,
            original_uuid=original_uuid,
            is_independent=is_independent,
        )
        self.project = self.run.project
        self.schedule_at = self.run.schedule_at
        self.started_at = self.run.started_at
        self.finished_at = self.run.finished_at
        self.duration = self.run.duration
        self.owner = self.project.owner

    @staticmethod
    def _collect_hub_refs(ops: List[V1Operation], owner_name: str) -> Dict:
        return {}

    @staticmethod
    def _get_upstream_run_params(
        owner_id: int,
        run_name: str,
        run_id: int,
        params: Dict[str, Dict],
        edge_kind: V1RunEdgeKind,
    ):
        if not params:
            return []

        params = {p: V1Param.construct(**params[p]) for p in params}

        upstream_run_params_by_names = ops_params.get_upstream_run_params_by_names(
            params=params
        )
        run_uuids = {
            param_spec.param.entity_ref
            for param_specs in upstream_run_params_by_names.values()
            for param_spec in param_specs
        }
        runs_queryset = Models.Run.objects
        if settings.HAS_ORG_MANAGEMENT:
            runs_queryset = runs_queryset.filter(project__owner__id=owner_id)
        run_ids_by_uuids = {
            v["uuid"].hex: v
            for v in runs_queryset.filter(uuid__in=run_uuids).values(
                "id", "uuid", "meta_info"
            )
        }
        if len(run_uuids) != len(run_ids_by_uuids):
            raise ValueError(
                f"Operation {run_name} is requesting outputs access to a run(s) that does not exists: "
                f"{set(run_uuids) - set(run_ids_by_uuids)}"
            )
        for run_upstream in upstream_run_params_by_names:
            if edge_kind not in V1RunEdgeKind.to_set():
                edge_kind = V1RunEdgeKind.RUN
            yield Models.RunEdge(
                upstream_id=run_ids_by_uuids[run_upstream]["id"],
                downstream_id=run_id,
                kind=edge_kind,
                values={
                    param_spec.name: param_spec.param.value
                    for param_spec in upstream_run_params_by_names[run_upstream]
                },
            )

    @classmethod
    def _resolve_edges(cls, run: BaseRun):
        # We check if we need to resolve any missing run params
        run_edges = []
        try:
            run_edges.extend(
                cls._get_upstream_run_params(
                    owner_id=run.project.owner_id,
                    run_name=run.name,
                    run_id=run.id,
                    params=run.params,
                    edge_kind=run.meta_info.get(META_EDGE, V1RunEdgeKind.RUN),
                )
            )
        except ValueError as e:
            raise AccessNotAuthorized(e)
        if run_edges:
            upstream_ids = {r.upstream_id for r in run_edges}
            # We check if any of the runs does not exist on the current project
            if (
                Models.Run.objects.exclude(project_id=run.project_id)
                .filter(id__in=upstream_ids)
                .count()
                > 0
            ):
                raise AccessNotAuthorized(
                    f"Run {run.name} request access to runs outside of the current project."
                )
            # We check the edge that require updating/creating (handles resume use case)
            run_edges_by_runs = {(r.upstream_id, r.downstream_id): r for r in run_edges}
            run_edges_keys = list(run_edges_by_runs.keys())
            query = reduce(
                or_, (Q(upstream_id=k[0], downstream_id=k[1]) for k in run_edges_keys)
            )
            to_update = Models.RunEdge.objects.filter(query).only(
                "id",
                "upstream_id",
                "downstream_id",
                "kind",
                "values",
            )
            _to_update = {(m.upstream_id, m.downstream_id) for m in to_update}
            to_create = {m for m in run_edges_keys if m not in _to_update}
            if to_update:
                updated = []
                for m in to_update:
                    key = (m.upstream_id, m.downstream_id)
                    m.kind = run_edges_by_runs[key].kind
                    m.values = run_edges_by_runs[key].values
                    updated.append(m)
                Models.RunEdge.objects.bulk_update(updated, ["kind", "values"])
            if to_create:
                Models.RunEdge.objects.bulk_create(
                    [run_edges_by_runs[m] for m in to_create]
                )

        return run_edges

    @staticmethod
    def _get_edge_queryset(queryset):
        return queryset.prefetch_related("artifacts").only(
            "id",
            "inputs",
            "outputs",
            "uuid",
            "name",
            "status",
            "created_at",
            "schedule_at",
            "started_at",
            "finished_at",
            "wait_time",
            "duration",
            "cloning_kind",
            "artifacts__id",
            "artifacts__name",
            "artifacts__path",
        )

    @staticmethod
    def _get_edge_param_value(
        edge_v: Union[str, Dict],
        inputs: Dict,
        outputs: Dict,
        artifacts: Dict,
        owner: str,
        project_name,
        uuid: str,
        name: str,
        status: str,
        condition: Dict,
        created_at: datetime,
        schedule_at: datetime,
        started_at: datetime,
        finished_at: datetime,
        duration: int,
        cloning_kind: V1CloningKind,
    ):
        artifacts_path = ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(uuid)
        outputs_path = ctx_paths.CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT.format(uuid)

        e_outputs = outputs or {}
        e_inputs = inputs or {}
        e_artifacts = artifacts or {}
        for k, v in e_artifacts.items():
            if v is not None:
                e_artifacts[k] = get_run_lineage_paths(uuid, [v])[0]
        e_globals = {
            ctx_keys.UUID: uuid,
            ctx_keys.NAME: name,
            ctx_keys.STATUS: status,
            ctx_keys.CONDITION: condition,
            ctx_keys.OWNER_NAME: owner,
            ctx_keys.PROJECT_NAME: project_name,
            ctx_keys.PROJECT_UNIQUE_NAME: get_project_instance(owner, project_name),
            ctx_keys.RUN_INFO: get_run_instance(owner, project_name, uuid),
            ctx_keys.CONTEXT_PATH: ctx_paths.CONTEXT_ROOT,
            ctx_keys.ARTIFACTS_PATH: ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
            ctx_keys.RUN_ARTIFACTS_PATH: artifacts_path,
            ctx_keys.RUN_OUTPUTS_PATH: outputs_path,
            ctx_keys.CREATED_AT: created_at,
            ctx_keys.SCHEDULE_AT: schedule_at,
            ctx_keys.STARTED_AT: started_at,
            ctx_keys.FINISHED_AT: finished_at,
            ctx_keys.DURATION: duration,
            ctx_keys.CLONING_KIND: cloning_kind,
        }

        def from_outputs():
            if not entity_value:
                return e_outputs
            if entity_value in e_outputs:
                return e_outputs[entity_value]

        def from_inputs():
            if not entity_value:
                return e_inputs
            if entity_value in e_inputs:
                return e_inputs[entity_value]

        def from_artifacts():
            outputs_subpath = ctx_paths.CONTEXTS_OUTPUTS_SUBPATH_FORMAT.format(uuid)
            if not entity_value:
                return dict(
                    {ctx_keys.BASE: uuid, ctx_sections.OUTPUTS: outputs_subpath},
                    **e_artifacts,
                )
            if entity_value == ctx_keys.BASE:
                return uuid
            if entity_value == ctx_sections.OUTPUTS:
                return outputs_subpath
            if entity_value in e_artifacts:
                return e_artifacts[entity_value]

        def from_globals():
            if not entity_value:
                return e_globals
            if entity_value in e_globals:
                return e_globals[entity_value]

        if isinstance(edge_v, dict):
            param_value = {}
            if "files" in edge_v:
                param_value["files"] = [
                    "{}/{}".format(uuid, i)
                    for i in to_list(edge_v["files"], check_none=True)
                ]
            if "dirs" in edge_v:
                param_value["dirs"] = [
                    "{}/{}".format(uuid, i)
                    for i in to_list(edge_v["dirs"], check_none=True)
                ]

            return param_value, True

        entity_type = ctx_refs.get_entity_type(edge_v)
        entity_value = ctx_refs.get_entity_value(edge_v)
        if entity_type == ctx_sections.INPUTS_OUTPUTS:
            return {
                ctx_sections.INPUTS: e_inputs,
                ctx_sections.OUTPUTS: e_outputs,
                ctx_sections.ARTIFACTS: e_artifacts,
            }, False
        if entity_type == ctx_sections.OUTPUTS:
            param_value = from_outputs()
        elif entity_type == ctx_sections.INPUTS:
            param_value = from_inputs()
        elif entity_type == ctx_sections.GLOBALS:
            param_value = from_globals()
        elif entity_type == ctx_sections.ARTIFACTS:
            param_value = from_artifacts()
        else:
            param_value = None

        return param_value, False

    @classmethod
    def _resolve_params(cls, run: BaseRun) -> Dict:
        # Get params by name
        params = run.params or {}
        if not params:
            return params

        def populate_context_from_results():
            owner = run.project.owner.name
            project_name = run.project.name
            for edge in edges:
                edge_values = edge.values
                edge_io = upstream_runs.get(edge.upstream_id)
                # Check if the edge still exists, user might have deleted the upstream run
                if not edge_io:
                    continue
                for edge_k, edge_v in edge_values.items():
                    param_value, is_value_artifacts = cls._get_edge_param_value(
                        edge_v=edge_v,
                        inputs=edge_io.inputs,
                        outputs=edge_io.outputs,
                        artifacts={a.name: a.path for a in edge_io.artifacts.all()},
                        owner=owner,
                        project_name=project_name,
                        uuid=edge_io.uuid.hex,
                        name=edge_io.name,
                        status=edge_io.status,
                        condition=edge_io.get_last_condition(),
                        created_at=edge_io.created_at,
                        schedule_at=edge_io.schedule_at,
                        started_at=edge_io.started_at,
                        finished_at=edge_io.finished_at,
                        duration=edge_io.duration,
                        cloning_kind=edge_io.cloning_kind,
                    )
                    if param_value:
                        current_param = params.get(edge_k, {})
                        current_param.pop("ref", None)
                        current_param["value"] = param_value
                        params[edge_k] = current_param

        # We get all run edges's values and then we resolve against the inputs/outputs
        edges = run.upstream_edges.exclude(values__isnull=True)
        queryset_without_cache = cls._get_edge_queryset(
            run.upstream_runs.exclude(cloning_kind=V1CloningKind.CACHE)
        )
        # Handle cache hits
        op_to_original = {
            v["original_id"]: v["id"]
            for v in run.upstream_runs.filter(cloning_kind=V1CloningKind.CACHE).values(
                "id", "original_id"
            )
        }
        queryset_cached = cls._get_edge_queryset(
            Models.Run.objects.filter(id__in=op_to_original.keys())
        )
        queryset = queryset_without_cache.union(queryset_cached)
        upstream_runs = {}
        for r in queryset:
            if r.id in op_to_original:
                upstream_runs[op_to_original[r.id]] = r
            else:
                upstream_runs[r.id] = r
        populate_context_from_results()

        # We get the params from the op
        return params

    def _resolve_joins(self) -> Dict:
        joins = self.compiled_operation.joins
        offline_filter = RunsOfflineFilter()
        edge_values_by_ids = {}
        param_values = {}
        owner = self.run.project.owner.name
        project_name = self.run.project.name

        def resolve_join_runs(j: V1Join):
            edge_values = {p: j.params[p].value for p in j.params}
            try:
                limit = RunsOfflineFilter.positive_int(
                    j.limit or 5000, strict=True, cutoff=5000
                )
                offset = RunsOfflineFilter.positive_int(j.offset or 0, strict=False)
            except (ValueError, TypeError):
                raise PolyaxonCompilerError(
                    "Received a wrong join (limit/offset) specification. "
                    "Join: {}.".format(join.to_dict())
                )
            queryset = Models.Run.objects.filter(project_id=self.run.project_id)
            try:
                queryset = offline_filter.filter_join(queryset=queryset, join=j)
            except PQLException as e:
                raise PolyaxonCompilerError(
                    "Received a wrong join specification. "
                    "Join: {}. Error: {}".format(join.to_dict(), e)
                )

            queryset = self._get_edge_queryset(queryset)[offset : offset + limit]
            for run_values in queryset:
                if run_values.id not in edge_values_by_ids:
                    edge_values_by_ids[run_values.id] = {}
                edge_values_by_ids[run_values.id].update(edge_values)
                for edge_k, edge_v in edge_values.items():
                    param_value, is_value_artifacts = self._get_edge_param_value(
                        edge_v=edge_v,
                        inputs=run_values.inputs,
                        outputs=run_values.outputs,
                        artifacts={a.name: a.path for a in run_values.artifacts.all()},
                        owner=owner,
                        project_name=project_name,
                        uuid=run_values.uuid.hex,
                        name=run_values.name,
                        status=run_values.status,
                        condition=run_values.get_last_condition(),
                        created_at=run_values.created_at,
                        schedule_at=run_values.schedule_at,
                        started_at=run_values.started_at,
                        finished_at=run_values.finished_at,
                        duration=run_values.duration,
                        cloning_kind=run_values.cloning_kind,
                    )
                    if param_value:
                        if is_value_artifacts:
                            if edge_k not in param_values:
                                param_values[edge_k] = {"files": [], "dirs": []}
                            param_values[edge_k] = {
                                "files": param_values[edge_k]["files"]
                                + param_value.get("files", []),
                                "dirs": param_values[edge_k]["dirs"]
                                + param_value.get("dirs", []),
                            }
                        else:
                            if edge_k not in param_values:
                                param_values[edge_k] = []
                            param_values[edge_k].append(param_value)
                    else:
                        if edge_k not in param_values:
                            param_values[edge_k] = []
                        param_values[edge_k].append(param_value)

        for join in joins:
            resolve_join_runs(join)

        run_edges = []
        for run_id in edge_values_by_ids:
            run_edges.append(
                Models.RunEdge(
                    upstream_id=run_id,
                    downstream_id=self.run.id,
                    kind=V1RunEdgeKind.JOIN,
                    values=edge_values_by_ids[run_id],
                )
            )

        if run_edges:
            Models.RunEdge.objects.bulk_create(run_edges)

        # Create params
        params = {}
        for join in joins:
            for edge_k in join.params:
                current_param = join.params.get(edge_k)
                if not current_param:
                    continue
                if edge_k not in params:
                    params[edge_k] = {}
                params[edge_k]["value"] = param_values.get(edge_k)
                if current_param.context_only:
                    params[edge_k]["contextOnly"] = current_param.context_only
                if current_param.connection:
                    params[edge_k]["connection"] = current_param.connection
                if current_param.to_init:
                    params[edge_k]["toInit"] = current_param.to_init

        return params

    def _resolve_joins_params(self):
        if self.compiled_operation.joins:
            self.apply_params(should_be_resolved=False)
            self.compiled_operation.joins = [
                V1Join.read(
                    CompiledOperationSpecification.apply_section_contexts(
                        config=self.compiled_operation,
                        section=join.to_dict(),
                        param_spec=self.param_spec,
                    )
                )  # TODO: Use construct
                for join in self.compiled_operation.joins
            ]
            self.params.update(self._resolve_joins())

    def resolve_edges(self):
        self._resolve_edges(run=self.run)

    def resolve_params(self):
        self.params = self._resolve_params(run=self.run)
        self._resolve_joins_params()

    def resolve_io(self):
        if self.compiled_operation.inputs:
            self.run.inputs = {
                io.name: io.value for io in self.compiled_operation.inputs
            }
        if self.compiled_operation.outputs:
            self.run.outputs = {
                io.name: io.value for io in self.compiled_operation.outputs
            }

    def _get_meta_artifacts_presets(self) -> List:
        if not self.run.meta_info or META_COPY_ARTIFACTS not in self.run.meta_info:
            return []

        artifacts = self.run.meta_info.pop(META_COPY_ARTIFACTS)
        artifacts = V1ArtifactsType.read(artifacts)

        def get_relative_to_run_artifacts(v: str):
            paths = v.split("/")[1:]
            paths = ["{{ globals.run_artifacts_path }}"] + paths
            return "/".join(paths)

        # Populate all paths
        if artifacts.dirs:
            artifacts.dirs = [
                [d, get_relative_to_run_artifacts(d)] for d in artifacts.dirs
            ]
        if artifacts.files:
            artifacts.files = [
                [d, get_relative_to_run_artifacts(d)] for d in artifacts.files
            ]
        init = V1Init.construct(artifacts=artifacts)
        return [{"runPatch": {"init": [init.to_dict()]}}]

    def _get_meta_destination_image(self) -> Optional[str]:
        meta_info = self.run.meta_info or {}
        return meta_info.get(META_DESTINATION_IMAGE)

    def resolve_presets(self):
        for preset in self._get_meta_artifacts_presets():
            self.compiled_operation = CompiledOperationSpecification.apply_preset(
                config=self.compiled_operation, preset=preset
            )

    def _pre_persist_state(self):
        # Slow sync process
        if not self.run.controller_id and not self.run.pipeline_id:
            return
        self.compiled_operation.plugins = self.compiled_operation.plugins or V1Plugins()
        self.compiled_operation.plugins.sidecar = (
            self.compiled_operation.plugins.sidecar or V1PolyaxonSidecarContainer()
        )
        if self.compiled_operation.plugins.sidecar.sync_interval is None:
            self.compiled_operation.plugins.sidecar.sync_interval = 60

    def _persist_meta_info(self):
        # Handle concurrency
        if self.run.is_matrix and self.compiled_operation.matrix.concurrency:
            self.run.meta_info[
                META_CONCURRENCY
            ] = self.compiled_operation.matrix.concurrency
        elif self.run.is_dag and self.compiled_operation.run.concurrency:
            self.run.meta_info[
                META_CONCURRENCY
            ] = self.compiled_operation.run.concurrency

        # Handle early stopping
        if self.run.is_matrix and self.compiled_operation.matrix.early_stopping:
            self.run.meta_info[META_HAS_EARLY_STOPPING] = True
        elif self.run.is_dag and self.compiled_operation.run.early_stopping:
            self.run.meta_info[META_HAS_EARLY_STOPPING] = True
        # handle services path
        if (
            self.compiled_operation.is_service_run
            and self.compiled_operation.run.rewrite_path
        ):
            self.run.meta_info[META_REWRITE_PATH] = True
        # handle services visibility
        if (
            self.compiled_operation.is_service_run
            and self.compiled_operation.run.is_external
        ):
            self.run.meta_info[META_IS_EXTERNAL] = True
        # handle services ports
        if self.compiled_operation.is_service_run and self.compiled_operation.run.ports:
            self.run.meta_info[META_PORTS] = self.compiled_operation.run.ports

    def _persist_resources(self):
        if self.compiled_operation.has_pipeline:
            return False
        resources = self.compiled_operation.run.get_resources()
        self.run.memory = resources.memory or 0
        self.run.cpu = resources.cpu or 0
        self.run.gpu = resources.gpu or 0
        self.run.custom = resources.custom or 0
        return True

    def _resolve_artifacts_lineage_state(self):
        self.artifacts = self.artifacts or []
        # Check upload and add it as a lineage
        if self.run.meta_info and META_UPLOAD_ARTIFACTS in self.run.meta_info:
            artifacts_path = self.run.meta_info[META_UPLOAD_ARTIFACTS]
            if artifacts_path:
                upload_artifact = collect_lineage_artifacts_path(artifacts_path)
                self.artifacts.append(upload_artifact)
        if self.artifacts:
            set_artifacts(self.run, self.artifacts)

    def _check_approval(self):
        if self.compiled_operation.is_approved is False and self.run.pending is None:
            self.run.pending = V1RunPending.APPROVAL
            return True
        return False

    def _persist_agent(self) -> bool:
        return False

    def resolve_state(self):
        self.run.state = get_run_state(
            cache=self.compiled_operation.cache,
            inputs=self.compiled_operation.inputs,
            outputs=self.compiled_operation.outputs,
            contexts=self.compiled_operation.contexts,
            init=self.compiled_operation.run.get_all_init(),
            connections=self.compiled_operation.run.get_all_connections(),
            containers=self.compiled_operation.run.get_all_containers(),
            namespace=self.project.uuid,
            component_state=self.run.component_state,
        )

    def persist_state(self):
        update_fields = [
            "content",
            "state",
            "inputs",
            "outputs",
            "meta_info",
        ]
        self._pre_persist_state()
        self._persist_meta_info()
        self._persist_resources()
        if self._persist_agent():
            update_fields += ["agent", "queue", "namespace", "artifacts_store"]
        if self._check_approval():
            update_fields.append("pending")
        if self.compiled_operation.cost:
            self.run.cost = self.compiled_operation.cost
            self.compiled_operation.cost = None
            update_fields.append("cost")
        if self._persist_resources():
            update_fields += ["memory", "cpu", "gpu", "custom"]

        self.run.content = self.compiled_operation.to_json()
        self.run.save(update_fields=update_fields)
        self._resolve_artifacts_lineage_state()

    @classmethod
    def _create_operation_dag(
        cls, run: BaseRun, compiled_operation: V1CompiledOperation
    ) -> BaseRun:
        run_config = compiled_operation.run  # type: V1Dag
        has_jobs = False
        has_services = False
        has_dags = False
        has_matrices = False
        has_schedules = False

        def get_upstream_op_params(op_dag_spec: DagOpSpec):
            if not op_dag_spec.upstream:
                return []
            upstream_op_params_by_names = ops_params.get_upstream_op_params_by_names(
                params=op_dag_spec.op.params
            )
            statuses_by_refs = op_dag_spec.op.get_upstream_statuses_events(
                op_dag_spec.upstream
            )
            for op_upstream in op_dag_spec.upstream:
                run_edge = Models.RunEdge(
                    upstream_id=runs_by_names[op_upstream].related_instance,
                    downstream_id=runs_by_names[op_name].related_instance,
                    kind=V1RunEdgeKind.DAG,
                    statuses=statuses_by_refs.get(op_upstream),
                )
                if op_upstream in upstream_op_params_by_names:
                    run_edge.values = {
                        param_spec.name: param_spec.param.value
                        for param_spec in upstream_op_params_by_names[op_upstream]
                    }

                yield run_edge

        def update_pipeline_params():
            # Resolve pipeline params
            pipeline_params = ops_params.get_dag_params_by_names(params=op_spec.params)
            if pipeline_params:
                pipeline_inputs = {i.name: i for i in (compiled_operation.inputs or {})}
                pipeline_contexts = {
                    i.name: i for i in (compiled_operation.contexts or {})
                }
                for pipeline_param in pipeline_params[ctx_refs.DAG_ENTITY_REF]:
                    param = pipeline_param.param
                    if pipeline_param.param.entity_value in pipeline_inputs:
                        io = pipeline_inputs[param.entity_value]
                        param = V1Param.construct(
                            value=io.value,
                            to_init=param.to_init or io.to_init,
                            connection=param.connection or io.connection,
                            context_only=param.context_only,
                        )
                    elif pipeline_param.param.entity_value in pipeline_contexts:
                        io = pipeline_contexts[pipeline_param.param.entity_value]
                        param = V1Param.construct(
                            value=io.value,
                            to_init=param.to_init or io.to_init,
                            connection=param.connection or io.connection,
                            context_only=param.context_only,
                        )
                    elif pipeline_param.param.entity_type == ctx_sections.GLOBALS:
                        # handles uid, uuid, and id
                        if pipeline_param.param.entity_value in ctx_keys.UUID:
                            param = V1Param.construct(
                                value=run.uuid.hex,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.NAME:
                            param = V1Param.construct(
                                value=run.name,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.STATUS:
                            param = V1Param.construct(
                                value=run.status,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.CONDITION:
                            param = V1Param.construct(
                                value=run.get_last_condition(),
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.OWNER_NAME:
                            param = V1Param.construct(
                                value=run.project.owner.name,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.PROJECT_UUID:
                            param = V1Param.construct(
                                value=run.project.uuid.hex,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.PROJECT_NAME:
                            param = V1Param.construct(
                                value=run.project.name,
                                context_only=param.context_only,
                            )
                        elif (
                            pipeline_param.param.entity_value
                            == ctx_keys.PROJECT_UNIQUE_NAME
                        ):
                            param = V1Param.construct(
                                value=get_project_instance(
                                    run.project.owner.name, run.project.name
                                ),
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.RUN_INFO:
                            param = V1Param.construct(
                                value=get_run_instance(
                                    run.project.owner.name,
                                    run.project.name,
                                    run.uuid.hex,
                                ),
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.CONTEXT_PATH:
                            param = V1Param.construct(
                                value=ctx_paths.CONTEXT_ROOT,
                                context_only=param.context_only,
                            )
                        elif (
                            pipeline_param.param.entity_value == ctx_keys.ARTIFACTS_PATH
                        ):
                            param = V1Param.construct(
                                value=ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
                                context_only=param.context_only,
                            )
                        elif (
                            pipeline_param.param.entity_value
                            == ctx_keys.RUN_ARTIFACTS_PATH
                        ):
                            param = V1Param.construct(
                                value=ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(
                                    run.uuid.hex
                                ),
                                context_only=param.context_only,
                            )
                        elif (
                            pipeline_param.param.entity_value
                            == ctx_keys.RUN_OUTPUTS_PATH
                        ):
                            param = V1Param.construct(
                                value=ctx_paths.CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT.format(
                                    run.uuid.hex
                                ),
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.CREATED_AT:
                            param = V1Param.construct(
                                value=run.created_at,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.SCHEDULE_AT:
                            param = V1Param.construct(
                                value=run.schedule_at,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.STARTED_AT:
                            param = V1Param.construct(
                                value=run.started_at,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.FINISHED_AT:
                            param = V1Param.construct(
                                value=run.finished_at,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.DURATION:
                            param = V1Param.construct(
                                value=run.duration,
                                context_only=param.context_only,
                            )
                        elif pipeline_param.param.entity_value == ctx_keys.CLONING_KIND:
                            param = V1Param.construct(
                                value=run.cloning_kind,
                                context_only=param.context_only,
                            )

                    else:
                        param = None
                    if param:
                        op_spec.params[pipeline_param.name] = param

        hub_refs_to_components = cls._collect_hub_refs(
            ops=run_config.operations, owner_name=run.project.owner.name
        )
        topological_sort = run_config.sort_topologically()
        # We create ops for each stage
        runs_by_names = {}
        tags_by_names = {}
        pipeline_override = {"patchStrategy": PatchStrategy.PRE_MERGE}
        if compiled_operation.cache:
            pipeline_override["cache"] = compiled_operation.cache.to_dict()
        if run_config.environment:
            pipeline_override["runPatch"] = {}
            pipeline_override["runPatch"][
                "environment"
            ] = run_config.environment.to_dict()
        if compiled_operation.termination:
            pipeline_override["termination"] = compiled_operation.termination.to_dict()
        if compiled_operation.plugins:
            pipeline_override["plugins"] = compiled_operation.plugins.to_dict()
        if compiled_operation.queue:
            pipeline_override["queue"] = compiled_operation.queue
        for stage in topological_sort:
            for op_name in stage:
                component_state = None
                if op_name in run_config.dag:
                    op_spec = run_config.dag[op_name]
                    if (
                        not op_spec.op.has_component_reference
                        and op_spec.op.has_hub_reference
                        and op_spec.op.hub_ref in hub_refs_to_components
                    ):
                        op_spec.op.set_definition(
                            hub_refs_to_components[op_spec.op.hub_ref][
                                "component_content"
                            ]
                        )
                        component_state = hub_refs_to_components[op_spec.op.hub_ref][
                            "component_state"
                        ]
                run_config.set_op_component(op_name)
                op_spec = run_config.get_op_spec_by_name(op_name)
                update_pipeline_params()
                runs_by_names[op_name] = operations.init_run(
                    project_id=run.project_id,
                    user_id=run.user_id,
                    pipeline_id=run.id,
                    op_spec=op_spec,
                    controller_id=run.controller_id or run.id,
                    override=pipeline_override,
                    supported_owners={run.project.owner.name},
                    component_state=component_state,
                )

                # Add events trigger flags
                if run_config.dag[op_spec.name].downstream:
                    for downstream_op in run_config.dag[op_spec.name].downstream:
                        if run_config.dag[downstream_op].op.has_events_for_upstream(
                            op_spec.name
                        ):
                            runs_by_names[op_name].instance.meta_info[
                                META_HAS_DOWNSTREAM_EVENTS_TRIGGER
                            ] = True
                            break

                # Add meta information
                if not has_jobs and runs_by_names[op_name].instance.is_job:
                    has_jobs = True
                if not has_services and runs_by_names[op_name].instance.is_service:
                    has_services = True
                if not has_dags and runs_by_names[op_name].instance.is_dag:
                    has_dags = True
                if not has_matrices and runs_by_names[op_name].instance.is_matrix:
                    has_matrices = True
                if not has_schedules and runs_by_names[op_name].instance.is_schedule:
                    has_schedules = True

                # Setting tags by op names
                tags = op_spec.tags or op_spec.definition.tags
                if tags:
                    tags_by_names[op_name] = tags

        # Create batch runs
        runs = Models.Run.objects.bulk_create(
            [runs_by_names[op_name].instance for op_name in runs_by_names]
        )
        # Extend tuple by run ids
        for pipeline_run in runs:
            runs_by_names[pipeline_run.name] = runs_by_names[pipeline_run.name].update(
                related_instance=pipeline_run.id,
            )

        run_edges = []
        for op_name in run_config.dag:
            run_edges.extend(get_upstream_op_params(run_config.dag[op_name]))

        if run_edges:
            Models.RunEdge.objects.bulk_create(run_edges)

        # Add meta to pipeline
        if has_jobs:
            run.meta_info[META_HAS_JOBS] = has_jobs
        if has_services:
            run.meta_info[META_HAS_SERVICES] = has_services
        if has_dags:
            run.meta_info[META_HAS_DAGS] = has_dags
        if has_matrices:
            run.meta_info[META_HAS_MATRICES] = has_matrices
        if has_schedules:
            run.meta_info[META_HAS_SCHEDULES] = has_schedules

        return run

    @staticmethod
    def _normalize_cron(cron: str) -> str:
        cron_presets = {
            "@hourly": "0 * * * *",
            "@daily": "0 0 * * *",
            "@weekly": "0 0 * * 0",
            "@monthly": "0 0 1 * *",
            "@quarterly": "0 0 1 */3 *",
            "@yearly": "0 0 1 1 *",
        }

        return cron_presets.get(cron, cron)

    @staticmethod
    def _is_schedule_done(
        num_runs: int,
        compiled_operation: V1CompiledOperation,
    ) -> bool:
        if (
            compiled_operation.schedule.max_runs
            and num_runs >= compiled_operation.schedule.max_runs
        ):
            return True
        if not compiled_operation.schedule.end_at:
            return False
        return now() >= compiled_operation.schedule.end_at

    @staticmethod
    def _get_next_from_interval(
        frequency: timedelta, start_at: datetime, last_start_at: datetime
    ) -> datetime:
        num_skips = 0
        if last_start_at:
            num_skips = (
                last_start_at - start_at
            ).total_seconds() / frequency.total_seconds()

        # if the last_start_at date is before the start date, we jump to the start date
        if num_skips < 0:
            num_skips = 0
        # if the `last_start_at` date falls exactly on an interval, jump to the next interval
        elif int(num_skips) == num_skips:
            num_skips += 1
        # otherwise jump to the next integer interval
        else:
            num_skips = int(num_skips + 1)

        interval = frequency * num_skips

        return start_at + interval

    @classmethod
    def _get_next_from_cron(
        cls, cron: str, start_at: datetime, last_start_at: datetime
    ) -> datetime:
        cron = cls._normalize_cron(cron)
        cron = croniter(cron, last_start_at or start_at)
        return cron.get_next(datetime)

    @classmethod
    def _get_next(
        cls, num_runs: int, run: BaseRun, compiled_operation: V1CompiledOperation
    ) -> datetime:
        if compiled_operation.has_datetime_schedule:
            return compiled_operation.schedule.start_at

        start_at = compiled_operation.schedule.start_at
        if num_runs == 0 and start_at:
            return start_at

        start_at = compiled_operation.schedule.start_at or run.created_at
        last_start_at = None
        if num_runs > 0:
            last_run = (
                run.pipeline_runs.order_by("created_at")
                .only("id", "schedule_at")
                .last()
            )
            if last_run and last_run.schedule_at:
                last_start_at = last_run.schedule_at
            else:
                last_start_at = now()
        if compiled_operation.has_interval_schedule:
            return cls._get_next_from_interval(
                frequency=compiled_operation.schedule.frequency,
                start_at=start_at,
                last_start_at=last_start_at,
            )
        elif compiled_operation.has_cron_schedule:
            return cls._get_next_from_cron(
                cron=compiled_operation.schedule.cron,
                start_at=start_at,
                last_start_at=last_start_at,
            )
        else:
            raise ValueError(
                "I should not be here, schedule not handled {}".format(
                    compiled_operation.schedule.kind
                )
            )

    @classmethod
    def _create_operation_schedule(
        cls,
        run: BaseRun,
        compiled_operation: V1CompiledOperation,
    ) -> bool:
        num_runs = run.pipeline_runs.count()
        success_condition = V1StatusCondition.get_condition(
            type=V1Statuses.SUCCEEDED,
            status="True",
            reason="PolyaxonScheduleFinished",
            message="Schedule finished.",
        )
        if cls._is_schedule_done(
            compiled_operation=compiled_operation, num_runs=num_runs
        ):
            new_run_status(run=run, condition=success_condition)
            return False

        start_at = cls._get_next(
            num_runs=num_runs, run=run, compiled_operation=compiled_operation
        )
        if (
            compiled_operation.schedule.end_at
            and start_at > compiled_operation.schedule.end_at
        ):
            new_run_status(run=run, condition=success_condition)
            return False

        op_spec = get_op_from_schedule(
            content=run.raw_content,
            compiled_operation=compiled_operation,
        )
        schedule_run = operations.init_run(
            project_id=run.project_id,
            user_id=run.user_id,
            pipeline_id=run.id,
            name=run.name,
            description=run.description,
            tags=run.tags,
            readme=run.readme,
            op_spec=op_spec,
            schedule_at=start_at,
            supported_owners={run.project.owner.name},
            component_state=run.component_state,
        ).instance
        # Create batch runs to avoid sending any signal yet
        Models.Run.objects.bulk_create([schedule_run])
        return True

    @staticmethod
    def _get_runs_from_ops(
        run: BaseRun,
        compiled_operation: V1CompiledOperation,
        suggestions: List[Dict],
        iteration: Optional[int] = None,
    ) -> List[BaseRun]:
        ops = get_ops_from_suggestions(
            content=run.raw_content,
            compiled_operation=compiled_operation,
            suggestions=suggestions,
        )

        date_now = now()
        meta_info = {}
        if META_UPLOAD_ARTIFACTS in run.meta_info:
            meta_info[META_COPY_ARTIFACTS] = V1ArtifactsType(
                dirs=[run.uuid.hex]
            ).to_dict()
            meta_info[META_UPLOAD_ARTIFACTS] = run.meta_info[META_UPLOAD_ARTIFACTS]
        component_state = None
        for op_spec in ops:
            # We make sure that the component state resolves to the correct runs' state
            # Calculate this value once
            if not component_state:
                component_state = get_component_version_state(op_spec.component)
            op_run = operations.init_run(
                project_id=run.project_id,
                user_id=run.user_id,
                pipeline_id=run.id,
                name=run.name,
                description=run.description,
                tags=run.tags,
                readme=run.readme,
                op_spec=op_spec,
                controller_id=run.controller_id or run.id,
                iteration=iteration,
                supported_owners={run.project.owner.name},
                component_state=component_state,
                meta_info=meta_info,
            ).instance
            op_run.uuid = uuid.uuid4().hex
            op_run.created_at = date_now
            op_run.updated_at = date_now
            yield op_run

    @classmethod
    def _create_mapping_matrix(
        cls, run: BaseRun, compiled_operation: V1CompiledOperation
    ):
        suggestions = MappingManager(compiled_operation.matrix).get_suggestions()
        runs = cls._get_runs_from_ops(
            run=run, compiled_operation=compiled_operation, suggestions=suggestions
        )
        # Create batch runs
        bulk_create_runs(runs)

    @classmethod
    def _create_grid_matrix(cls, run: BaseRun, compiled_operation: V1CompiledOperation):
        suggestions = GridSearchManager(compiled_operation.matrix).get_suggestions()
        runs = cls._get_runs_from_ops(
            run=run, compiled_operation=compiled_operation, suggestions=suggestions
        )
        # Create batch runs
        bulk_create_runs(runs)

    @classmethod
    def _create_random_matrix(
        cls, run: BaseRun, compiled_operation: V1CompiledOperation
    ):
        suggestions = RandomSearchManager(compiled_operation.matrix).get_suggestions()
        runs = cls._get_runs_from_ops(
            run=run, compiled_operation=compiled_operation, suggestions=suggestions
        )
        # Create batch runs
        bulk_create_runs(runs)

    @classmethod
    def _create_operation_matrix(
        cls, run: BaseRun, compiled_operation: V1CompiledOperation
    ) -> None:
        if compiled_operation.has_random_search_matrix:
            cls._create_random_matrix(run=run, compiled_operation=compiled_operation)
        elif compiled_operation.has_grid_search_matrix:
            cls._create_grid_matrix(run=run, compiled_operation=compiled_operation)
        elif compiled_operation.has_mapping_matrix:
            cls._create_mapping_matrix(run=run, compiled_operation=compiled_operation)

    @staticmethod
    def _get_pipeline_join(
        pipeline: str,
        metric: V1OptimizationMetric,
        limit: int = 1000,
        additional_filters: Optional[str] = None,
    ):
        query = "pipeline:{}, metrics.{}:~nil, status: {}".format(
            pipeline, metric.name, V1Statuses.SUCCEEDED
        )
        if additional_filters:
            query = "{}, {}".format(query, additional_filters)

        return V1Join.construct(
            query=query,
            sort=metric.get_for_sort(),
            limit=limit,
            params={
                "configs": V1JoinParam.construct(value="inputs"),
                "metrics": V1JoinParam.construct(
                    value="outputs.{}".format(metric.name)
                ),
            },
        )

    @classmethod
    def _create_iteration_matrix(
        cls,
        run: BaseRun,
        suggestions: List[Dict],
        iteration: int,
        compiled_operation: V1CompiledOperation,
    ) -> Optional[List[int]]:
        runs = cls._get_runs_from_ops(
            run=run,
            compiled_operation=compiled_operation,
            suggestions=suggestions,
            iteration=iteration,
        )
        # Create batch runs
        if runs:
            runs = Models.Run.objects.bulk_create(runs)
            return [r.id for r in runs]
        else:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.SUCCEEDED,
                status="True",
                reason="SchedulerEmptySuggestions",
                message="Iteration algorithm received an empty suggestion list.",
            )
            new_run_status(run=run, condition=condition)
            return False

    def _apply_pipeline_contexts(self):
        if self.compiled_operation.schedule:
            self._create_operation_schedule(
                run=self.run, compiled_operation=self.compiled_operation
            )
        elif self.compiled_operation.matrix:
            self._create_operation_matrix(
                run=self.run, compiled_operation=self.compiled_operation
            )
        elif self.compiled_operation.is_dag_run:
            self.run = self._create_operation_dag(
                run=self.run, compiled_operation=self.compiled_operation
            )
        return self.compiled_operation

    @staticmethod
    def _get_cache_query(
        run, filters: Optional[Dict] = None, excludes: Optional[Dict] = None
    ):
        queryset = Models.Run.objects.filter(
            state=run.state, kind=run.kind, runtime=run.runtime
        ).exclude(id=run.id)
        if filters:
            queryset = queryset.filter(**filters)
        if excludes:
            queryset = queryset.exclude(**excludes)

        return queryset.order_by("created_at").only("id", "original").last()

    @classmethod
    def _is_cache_hit(
        cls, run: BaseRun, compiled_operation: V1CompiledOperation
    ) -> bool:
        # Check if the run was approved to be scheduled
        if run.original_id and run.cloning_kind == V1CloningKind.CACHE:
            run.cloning_kind = None
            run.original_id = None
            run.pending = None
            run.save(
                update_fields=["original_id", "cloning_kind", "updated_at", "pending"]
            )
            return False
        if compiled_operation.cache and compiled_operation.cache.disable:
            return False

        if not compiled_operation.cache or compiled_operation.cache.disable is None:
            # Cache is disabled by default for independent ops
            if not run.pipeline_id:
                return False
            # Cache is disabled by default for schedules
            if run.is_schedule:
                return False
        # Cache is disabled by default for cloned ops
        if run.original_id:
            return False
        # User did not define a state to check
        if not run.state:
            return False

        cached = cls._get_cache_query(
            run,
            filters={
                "status__in": {
                    V1Statuses.PROCESSING,
                    V1Statuses.SCHEDULED,
                    V1Statuses.STARTING,
                    V1Statuses.RUNNING,
                    V1Statuses.SUCCEEDED,
                }
            },
        )

        # Not cached check all runs from the same queue
        if not cached:
            cached = cls._get_cache_query(
                run,
                excludes={
                    "status__in": {
                        V1Statuses.FAILED,
                        V1Statuses.SKIPPED,
                        V1Statuses.UPSTREAM_FAILED,
                        V1Statuses.STOPPED,
                    }
                },
            )

        # No cache found
        if not cached:
            return False

        # Check if cache has expired
        cached = cached.original if cached.original else cached
        if (
            compiled_operation.cache
            and compiled_operation.cache.ttl is not None
            and cached.finished_at
        ):
            if (
                now() - cached.finished_at
            ).total_seconds() >= compiled_operation.cache.ttl:
                return False

        # Check if cache has a stopping triggered
        if LifeCycle.is_stopping(cached.status):
            return False

        # Cache finished but not successfully
        if LifeCycle.is_done(cached.status) and not LifeCycle.succeeded(cached.status):
            return False

        # Use cache
        run.original_id = cached.id
        run.cloning_kind = V1CloningKind.CACHE
        if LifeCycle.succeeded(cached.status):
            run.outputs = cached.outputs
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.SUCCEEDED,
                status="True",
                reason="CacheHit",
                message="Run is using cached results",
            )
            new_run_status(
                run=run,
                condition=condition,
                additional_fields=["original_id", "cloning_kind", "outputs"],
            )
        else:
            run.pending = V1RunPending.CACHE
            run.save(
                update_fields=["original_id", "cloning_kind", "updated_at", "pending"]
            )
        return True

    @staticmethod
    def _validation_conditions(
        run: BaseRun,
        conditions: Any = None,
    ) -> bool:
        if conditions is None:
            return True

        try:
            conditions = to_bool(conditions)
        except Exception as e:
            raise PolyaxonCompilerError(
                "Received a wrong expression for the key: conditions. "
                "The expression could be evaluated to a boolean value. "
                "Condition: {}. Error: {}".format(conditions, e)
            )

        if conditions:
            return True

        condition = V1StatusCondition.get_condition(
            type=V1Statuses.SKIPPED,
            status="True",
            reason="ConditionsCheckFailed",
            message="The 'conditions' field was evaluated to False.",
        )
        new_run_status(run=run, condition=condition)
        return False

    @staticmethod
    def _validation_hook_conditions(
        hook_index: int,
        conditions: Any = None,
    ) -> bool:
        if conditions is None:
            return True

        try:
            conditions = to_bool(conditions)
        except Exception as e:
            raise PolyaxonCompilerError(
                "Hook `{}` received a wrong expression for the key: conditions. "
                "The expression could be evaluated to a boolean value. "
                "Condition: {}. Error: {}".format(hook_index, conditions, e)
            )

        return conditions

    def _should_skip_runtime_resolution(self):
        # Check conditions
        if not self._validation_conditions(
            run=self.run, conditions=self.compiled_operation.conditions
        ):
            return True
        # Check cache
        return self._is_cache_hit(
            run=self.run, compiled_operation=self.compiled_operation
        )

    def _get_project_version_refs(
        self,
        names: Union[List[str], Set[str]],
        kind: V1ProjectVersionKind,
    ) -> List[BaseProjectVersion]:
        names = set(names)
        entities = []
        query = []
        for name in names:
            try:
                owner, entity_namespace, version = get_versioned_entity_info(
                    entity=name,
                    entity_name="{} version".format(kind),
                    default_owner=self.owner_name,
                )
                entities.append((owner, entity_namespace, version))
                query.append(
                    Q(
                        project__owner__name=owner,
                        project__name=entity_namespace,
                        name=version,
                    )
                    if settings.HAS_ORG_MANAGEMENT
                    else Q(
                        project__name=entity_namespace,
                        name=version,
                    )
                )
            except Exception as e:
                raise AccessNotFound(
                    "A {} version reference was provided "
                    "but could not be parsed correctly: `{}`. "
                    "Error: {}".format(kind, name, e)
                )
        versions = Models.ProjectVersion.objects.filter(kind=kind)
        if settings.HAS_ORG_MANAGEMENT:
            versions = versions.filter(project__owner__name=self.owner_name)
        versions = versions.filter(reduce(or_, query))
        additional_fields = ()
        if settings.HAS_ORG_MANAGEMENT:
            versions.select_related("connection")
            additional_fields = (
                "connection_id",
                "connection__name",
            )
        versions = (
            versions.prefetch_related("lineage", "lineage__artifact")
            .annotate(lineage_count=Count("lineage"))
            .filter(lineage_count__gt=0)
        )
        versions = versions.only(
            "id",
            "project_id",
            "lineage__id",
            "lineage__artifact__id",
            "lineage__artifact__name",
            "lineage__artifact__path",
            "lineage__artifact__kind",
            "run_id",
            "run__uuid",
            *additional_fields,
        )
        ids = {v.id for v in versions}
        # connection_ids = {v.connection_id for v in versions}
        # project_ids = {
        #     v.project_id for v in versions if v.project_id != project_settings.id
        # }
        if len(ids) != len(names):
            raise AccessNotFound(
                "Some {} version refs were provided "
                "but were not found: `{}`".format(kind, names)
            )

        # if (
        #     project_ids
        #     and project_settings
        #     and project_settings.projects.exists()
        #     and project_settings.projects.filter(id__in=project_ids).count()
        #     != len(project_ids)
        # ):
        #     raise AccessNotAuthorized(
        #         "Some {} versions were provided but are not authorised on this project.".format(
        #             kind
        #         )
        #     )
        # if (
        #     connection_ids
        #     and project_settings
        #     and project_settings.connections.exists()
        #     and project_settings.connections.filter(id__in=connection_ids).count()
        #     != len(connection_ids)
        # ):
        #     raise AccessNotAuthorized(
        #         "Some {} versions were provided but they are stored on connections not authorised on this project.".format(
        #             kind
        #         )
        #     )
        return versions

    def _get_project_version_paths(
        self,
        version_names: Set[str],
        kind: V1ProjectVersionKind,
    ) -> Tuple[List[V1Init], Set[str]]:
        refs = self._get_project_version_refs(
            names=version_names,
            kind=kind,
        )

        init = []
        connections = set([])
        for r in refs:
            connection_name = r.connection.name if getattr(r, "connection_id") else None
            if connection_name:
                connections.add(connection_name)
            paths = get_run_lineage_paths(
                r.run.uuid.hex,
                r.lineage.values_list("artifact__path", flat=True),
            )
            if paths:
                init.append(
                    V1Init.construct(
                        connection=connection_name,
                        paths=paths,
                    )
                )
        return init, connections

    def _get_run_refs(
        self,
        names: Union[List[str], Set[str]],
    ) -> List[List[str]]:
        names = set(names)
        entities = []
        query = []
        lineages_by_run = {}
        for name in names:
            try:
                owner, entity_namespace, lineage = get_versioned_entity_info(
                    entity=name,
                    entity_name="lineage",
                    default_owner=self.owner_name,
                )
                entities.append((owner, entity_namespace, lineage))
                if entity_namespace not in lineages_by_run:
                    lineages_by_run[entity_namespace] = []
                lineages_by_run[entity_namespace].append(lineage)
                query.append(
                    Q(
                        project__owner__name=owner,
                        uuid=entity_namespace,
                    )
                    if settings.HAS_ORG_MANAGEMENT
                    else Q(uuid=entity_namespace)
                )
            except Exception as e:
                raise AccessNotFound(
                    "A lineage reference was provided "
                    "but could not be parsed correctly: `{}`. "
                    "Error: {}".format(name, e)
                )
        runs = Models.Run.objects
        if settings.HAS_ORG_MANAGEMENT:
            runs = runs.filter(project__owner__name=self.owner_name)
        runs = (
            runs.filter(reduce(or_, query))
            .prefetch_related("artifacts")
            .annotate(artifacts_count=Count("artifacts"))
            .filter(artifacts_count__gt=0)
        )
        runs = runs.only(
            "id",
            "uuid",
            "project_id",
            "artifacts__id",
            "artifacts__name",
            "artifacts__path",
            "artifacts__kind",
        )
        ids = {v.id for v in runs}
        # connection_ids = {v.connection_id for v in versions}
        # project_ids = {
        #     v.project_id for v in versions if v.project_id != project_settings.id
        # }
        lineage_runs = list(lineages_by_run.keys())
        if len(ids) != len(lineage_runs):
            raise AccessNotFound(
                "Some lineage refs were provided but were not found. "
                "Please check that the runs `{}` exist and are available".format(
                    lineage_runs
                )
            )

        # if (
        #     project_ids
        #     and project_settings
        #     and project_settings.projects.exists()
        #     and project_settings.projects.filter(id__in=project_ids).count()
        #     != len(project_ids)
        # ):
        #     raise AccessNotAuthorized(
        #         "Some {} versions were provided but are not authorised on this project.".format(
        #             kind
        #         )
        #     )
        # if (
        #     connection_ids
        #     and project_settings
        #     and project_settings.connections.exists()
        #     and project_settings.connections.filter(id__in=connection_ids).count()
        #     != len(connection_ids)
        # ):
        #     raise AccessNotAuthorized(
        #         "Some {} versions were provided but they are stored on connections not authorised on this project.".format(
        #             kind
        #         )
        #     )
        return [
            get_run_lineage_paths(
                r.uuid.hex,
                r.artifacts.filter(name__in=lineages_by_run[r.uuid.hex]).values_list(
                    "path", flat=True
                ),
            )
            for r in runs
        ]

    def _get_lineage_init_paths(self) -> Tuple[List[V1Init], Set[str]]:
        lineage_names = CompiledOperationSpecification.get_init_lineage_refs(
            config=self.compiled_operation,
        )
        if not lineage_names:
            return [], set([])

        refs = self._get_run_refs(names=lineage_names)

        init = []
        connections = set([])
        for r in refs:
            # connection_name = r.connection.name if r.connection_id else None
            # if connection_name:
            #     connections.add(connection_name)
            if r:
                init.append(
                    V1Init.construct(
                        # connection=connection_name,
                        paths=r,
                    )
                )
        return init, connections

    def _get_model_version_paths(
        self,
    ) -> Tuple[List[V1Init], Set[str]]:
        version_names = CompiledOperationSpecification.get_init_model_refs(
            config=self.compiled_operation,
        )
        if not version_names:
            return [], set([])

        return self._get_project_version_paths(
            version_names=version_names,
            kind=V1ProjectVersionKind.MODEL,
        )

    def _get_artifact_version_paths(self) -> Tuple[List[V1Init], Set[str]]:
        version_names = CompiledOperationSpecification.get_init_artifact_refs(
            config=self.compiled_operation,
        )
        if not version_names:
            return [], set([])

        return self._get_project_version_paths(
            version_names=version_names,
            kind=V1ProjectVersionKind.ARTIFACT,
        )

    def resolve_init_refs(self):
        m_init, _ = self._get_model_version_paths()
        a_init, _ = self._get_artifact_version_paths()
        l_init, _ = self._get_lineage_init_paths()
        init = m_init + a_init + l_init
        if not init:
            return
        init = [i.to_dict() for i in init]
        model_preset = {"runPatch": {"init": init}}

        self.compiled_operation = CompiledOperationSpecification.apply_preset(
            config=self.compiled_operation, preset=model_preset
        )

    def clean_init_refs(self):
        self.compiled_operation = (
            CompiledOperationSpecification.clean_init_version_refs(
                self.compiled_operation
            )
        )

    def resolve_build(self):
        if not self.compiled_operation.build:
            return

        inputs = {i.name: i.value for i in self.compiled_operation.inputs or {}}
        build_instance = operations.resolve_build(
            project_id=self.run.project_id,
            user_id=self.run.user_id,
            compiled_operation=self.compiled_operation,
            inputs=inputs,
            meta_info=None,
            pending=None,
            supported_owners={self.owner_name},
        )
        # Change the pending logic to wait for the build
        self.compiled_operation.build = None
        self.run.pending = V1RunPending.BUILD
        self.run.content = self.compiled_operation.to_json()
        self.run.save(update_fields=["content", "updated_at", "pending"])

        operations.save_build_relation(
            operations.OperationInitSpec(
                compiled_operation=None,
                instance=self.run,
                related_instance=build_instance,
            )
        )
        return self.compiled_operation

    def apply_hooks_contexts(self):
        contexts = self._resolve_contexts()
        return CompiledOperationSpecification.apply_hooks_contexts(
            self.compiled_operation, contexts=contexts
        )

    def resolve_hooks(self) -> List[V1Operation]:
        self.resolve_connections()
        hooks = self.apply_hooks_contexts()
        ops = []
        contexts = dict(
            name=self.run.name,
            uuid=self.run.uuid.hex,
            kind=self.run.kind,
            status=self.run.status,
            runtime=self.run.runtime,
            wait_time=self.run.wait_time,
            duration=self.run.duration,
            inputs=self.run.inputs,
            outputs=self.run.outputs,
            condition=self.run.get_last_condition(),
        )
        for i, hook in enumerate(hooks):
            if hook.check_trigger_for_status(
                status=self.run.status
            ) and self._validation_hook_conditions(
                hook_index=i, conditions=hook.conditions
            ):
                ops.append(
                    V1Operation.from_hook(
                        hook,
                        contexts=contexts,
                    ),
                )
        return ops



from django.conf import settings

from haupt.apis.endpoints.project import ProjectResourceListEndpoint
from haupt.apis.methods import project_resources as methods
from haupt.apis.serializers.artifacts import (
    RunArtifactLightSerializer,
    RunArtifactSerializer,
)
from haupt.apis.serializers.base.bookmarks_mixin import BookmarkedListMixinView
from haupt.apis.serializers.runs import (
    BookmarkedRunSerializer,
    BookmarkedTimelineRunSerializer,
    GraphRunSerializer,
    OfflineRunSerializer,
    OperationCreateSerializer,
)
from haupt.common.apis.filters import OrderingFilter, QueryFilter
from haupt.common.apis.paginator import LargeLimitOffsetPagination
from haupt.common.endpoints.base import (
    CreateEndpoint,
    DestroyEndpoint,
    ListEndpoint,
    PostEndpoint,
)
from haupt.db.defs import Models
from haupt.db.managers.flows import get_run_graph
from haupt.db.queries import artifacts as artifacts_queries
from haupt.db.queries.artifacts import clean_sqlite_distinct_artifacts
from haupt.db.queries.runs import DEFAULT_COLUMNS_DEFER
from haupt.db.query_managers.artifact import ArtifactQueryManager
from haupt.db.query_managers.run import RunQueryManager


class ProjectRunsTagView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.create_runs_tags(view=self, request=request, *args, **kwargs)


class ProjectRunsStopView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.stop_runs(
            view=self, request=request, actor=request.user, *args, **kwargs
        )


class ProjectRunsApproveView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.approve_runs(
            view=self, request=request, actor=request.user, *args, **kwargs
        )


class ProjectRunsInvalidateView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.invalidate_runs(
            view=self, request=request, actor=self.request.user, *args, **kwargs
        )


class ProjectRunsBookmarkView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.bookmarks_runs(
            view=self, request=request, actor=self.request.user, *args, **kwargs
        )


class ProjectRunsArchiveView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.archive_runs(
            view=self, request=request, actor=request.user, *args, **kwargs
        )


class ProjectRunsRestoreView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.restore_runs(
            view=self, request=request, actor=request.user, *args, **kwargs
        )


class ProjectRunsTransferView(ProjectResourceListEndpoint, PostEndpoint):
    ALLOWED_METHODS = ["POST"]

    def post(self, request, *args, **kwargs):
        return methods.transfer_runs(
            view=self, request=request, actor=request.user, *args, **kwargs
        )


class ProjectRunsDeleteView(ProjectResourceListEndpoint, DestroyEndpoint):
    ALLOWED_METHODS = ["DELETE"]

    def _pre_delete_check(self, request):
        pass

    def delete(self, request, *args, **kwargs):
        self._pre_delete_check(request)
        return methods.delete_runs(
            view=self, request=request, actor=self.request.user, *args, **kwargs
        )


class ProjectRunsListMixinView(BookmarkedListMixinView):
    bookmarked_model = "run"

    def handle_graph(self, serializer_class, *args, **kwargs):
        queryset = args[0]

        if not queryset:
            return super().get_serializer(*args, **kwargs)

        object_ids = {o.id for o in queryset}

        # Get graph
        graph = get_run_graph({"id__in": object_ids})

        context = self.get_serializer_context()
        context["graph"] = graph
        kwargs["context"] = context
        return serializer_class(*args, **kwargs)

    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class()
        if serializer_class == GraphRunSerializer:
            return self.handle_graph(serializer_class, *args, **kwargs)

        return super().get_serializer(*args, **kwargs)


class ProjectRunsListView(
    ProjectRunsListMixinView, ProjectResourceListEndpoint, ListEndpoint, CreateEndpoint
):
    queryset = Models.Run.all.prefetch_related("original", "pipeline", "user").defer(
        *DEFAULT_COLUMNS_DEFER
    )
    pagination_class = LargeLimitOffsetPagination
    query_manager = RunQueryManager
    check_alive = RunQueryManager.CHECK_ALIVE
    ordering = RunQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = RunQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = RunQueryManager.FIELDS_ORDERING_PROXY
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True
    ALLOWED_METHODS = ["GET", "POST"]
    serializer_class_mapping = {
        "GET": BookmarkedRunSerializer,
        "POST": OperationCreateSerializer,
    }

    def get_serializer_class(self):
        if self.request.method == "POST":
            return OperationCreateSerializer
        elif self.request.method == "GET":
            if self.request.query_params.get("mode") == "timeline":
                return BookmarkedTimelineRunSerializer
            elif self.request.query_params.get("mode") == "graph":
                return GraphRunSerializer
            else:
                return BookmarkedRunSerializer

        else:
            return super().get_serializer_class()

    def perform_create(self, serializer):
        serializer.save(project=self.project)


class ProjectRunsSyncView(ProjectResourceListEndpoint, CreateEndpoint):
    queryset = Models.Run.all.all()
    serializer_class = OfflineRunSerializer
    ALLOWED_METHODS = ["POST"]

    def perform_create(self, serializer):
        serializer.save(project=self.project)


class ProjectRunsArtifactsView(ProjectResourceListEndpoint, ListEndpoint):
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = ArtifactQueryManager
    check_alive = ArtifactQueryManager.CHECK_ALIVE
    ordering = ArtifactQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ArtifactQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ArtifactQueryManager.FIELDS_ORDERING_PROXY
    pagination_class = LargeLimitOffsetPagination
    ALLOWED_METHODS = ["GET"]

    def enrich_queryset(self, queryset):
        return queryset.filter(run__project=self.project)

    def get_queryset(self):
        mode = self.request.query_params.get("mode")
        if not mode:
            return artifacts_queries.project_runs_artifacts
        return artifacts_queries.project_runs_artifacts_distinct

    def get_serializer_class(self):
        mode = self.request.query_params.get("mode")
        if not mode:
            return RunArtifactSerializer
        return RunArtifactLightSerializer

    def list(self, request, *args, **kwargs):
        response = super().list(request, *args, **kwargs)
        mode = self.request.query_params.get("mode")
        if settings.DB_ENGINE_NAME == "sqlite" and mode:
            (
                response.data["results"],
                response.data["count"],
            ) = clean_sqlite_distinct_artifacts(response.data["results"])
            return response
        return response



from haupt.proxies.schemas.base import get_config
from polyaxon.api import AUTH_REQUEST_V1_LOCATION, K8S_V1_LOCATION

K8S_LOCATION_OPTIONS = r"""
location {app} {{
    auth_request     {auth_request};
    auth_request_set $auth_status $upstream_status;
    auth_request_set $k8s_token $upstream_http_k8s_token;
    auth_request_set $k8s_uri $upstream_http_k8s_uri;
    proxy_pass $k8s_uri;
    proxy_http_version 1.1;
    proxy_redirect     off;
    proxy_set_header Upgrade $http_upgrade;
    proxy_set_header Connection "upgrade";
    proxy_hide_header X-Frame-Options;
    proxy_set_header Origin "";
    proxy_set_header Host $http_host;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_set_header Authorization "bearer $k8s_token";
    proxy_buffering off;
}}
"""  # noqa


def get_k8s_root_location_config():
    return get_config(
        options=K8S_LOCATION_OPTIONS,
        app=K8S_V1_LOCATION,
        auth_request=AUTH_REQUEST_V1_LOCATION,
    )



import os

from typing import Dict, Optional

from clipped.utils.imports import import_string

from django.conf import settings
from django.db import transaction

from haupt.common import conf
from haupt.common.options.registry.core import SCHEDULER_ENABLED

if settings.SCHEDULER_ENABLED:
    from haupt.background.celeryp.app import app
    from haupt.background.celeryp.polyaxon_task import HauptTask

    app.Task = HauptTask  # Custom base class for logging


def send(
    task_name,
    delay: Optional[bool] = None,
    kwargs: Optional[Dict] = None,
    eager_kwargs: Optional[Dict] = None,
    **options,
):
    delay = conf.get(SCHEDULER_ENABLED) if delay is None else delay
    if not delay:
        module = os.environ.get("CONFIG_PREFIX", "haupt")
        tasks_execution = import_string(
            f"{module}.background.celeryp.executions.TasksExecutions"
        )
        tasks_execution.run(task=task_name, kwargs=kwargs, eager_kwargs=eager_kwargs)
        return
    options["ignore_result"] = options.get("ignore_result", True)
    return transaction.on_commit(
        lambda: app.send_task(task_name, kwargs=kwargs, **options)
    )
    # if "countdown" not in options:
    #     options["countdown"] = conf.get(SCHEDULER_GLOBAL_COUNTDOWN)
    # return app.send_task(task_name, kwargs=kwargs, **options)



from haupt.common.apis.regex import (
    ARTIFACT_NAME_PATTERN,
    OWNER_NAME_PATTERN,
    PROJECT_NAME_PATTERN,
    RUN_UUID_PATTERN,
)

URLS_RUNS_CREATE = r"^{}/{}/runs/create/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_RUNS_LIST = r"^{}/runs/list/?$".format(OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN)
URLS_RUNS_DETAILS = r"^{}/{}/runs/{}/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_RESTART = r"^{}/{}/runs/{}/restart/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_RESUME = r"^{}/{}/runs/{}/resume/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_COPY = r"^{}/{}/runs/{}/copy/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_STOP = r"^{}/{}/runs/{}/stop/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_APPROVE = r"^{}/{}/runs/{}/approve/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_STATUSES = r"^{}/{}/runs/{}/statuses/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_NAMESPACE = r"^{}/{}/runs/{}/namespace/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_ARTIFACTS_LINEAGE_LIST_V0 = r"^{}/{}/runs/{}/artifacts_lineage/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_ARTIFACTS_LINEAGE_NAMES_V0 = (
    r"^{}/{}/runs/{}/artifacts_lineage/names/?$".format(
        OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
    )
)
URLS_RUNS_ARTIFACTS_LINEAGE_DETAILS_V0 = (
    r"^{}/{}/runs/{}/artifacts_lineage/{}/?$".format(
        OWNER_NAME_PATTERN,
        PROJECT_NAME_PATTERN,
        RUN_UUID_PATTERN,
        ARTIFACT_NAME_PATTERN,
    )
)
URLS_RUNS_ARTIFACTS_LINEAGE_LIST = r"^{}/{}/runs/{}/lineage/artifacts/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_ARTIFACTS_LINEAGE_NAMES = r"^{}/{}/runs/{}/lineage/artifacts/names/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_ARTIFACTS_LINEAGE_DETAILS = r"^{}/{}/runs/{}/lineage/artifacts/{}/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN, ARTIFACT_NAME_PATTERN
)
URLS_RUNS_UPSTREAM_LINEAGE_LIST = r"^{}/{}/runs/{}/lineage/upstream/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_DOWNSTREAM_LINEAGE_LIST = r"^{}/{}/runs/{}/lineage/downstream/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_CLONES_LINEAGE_LIST = r"^{}/{}/runs/{}/lineage/clones/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)

URLS_RUNS_TRANSFER = r"^{}/{}/runs/{}/transfer/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_INVALIDATE = r"^{}/{}/runs/{}/invalidate/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_ARCHIVE = r"^{}/{}/runs/{}/archive/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_RESTORE = r"^{}/{}/runs/{}/restore/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)

URLS_RUNS_BOOKMARK = r"^{}/{}/runs/{}/bookmark/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)
URLS_RUNS_UNBOOKMARK = r"^{}/{}/runs/{}/unbookmark/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)

URLS_RUNS_STATS = r"^{}/{}/runs/{}/stats/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, RUN_UUID_PATTERN
)



from haupt.proxies.schemas.base import get_config

OPTIONS = """
location = /robots.txt {{
    rewrite ^ /static/robots.txt;
}}
"""


def get_robots_config():
    return get_config(options=OPTIONS, indent=0)



from functools import reduce
from operator import or_

from rest_framework import serializers

from django.conf import settings
from django.db.models import Q

from haupt.common import conf
from haupt.common.options.registry.k8s import K8S_NAMESPACE
from haupt.db.defs import Models
from polyaxon._constants.globals import DEFAULT_HUB
from polyaxon.schemas import V1ProjectVersionKind, V1RunEdgeKind, V1RunKind


class SettingsMixin:
    settings = serializers.SerializerMethodField()

    def get_settings(self, obj):
        return {
            "namespace": obj.namespace
            if hasattr(obj, "namespace")
            else conf.get(K8S_NAMESPACE),
            "component": {"state": obj.component_state.hex}
            if obj.component_state
            else None,
        }


class FullSettingsMixin:
    def get_settings(self, obj):
        tensorboard = None
        if obj.kind == V1RunKind.JOB:
            tensorboard = (
                obj.downstream_runs.filter(upstream_edges__kind=V1RunEdgeKind.TB)
                .order_by("created_at")
                .last()
            )
        build = (
            obj.upstream_runs.filter(downstream_edges__kind=V1RunEdgeKind.BUILD)
            .order_by("created_at")
            .last()
        )
        filters = [Q(run=obj)]
        if obj.component_state:
            filters.append(Q(state=obj.component_state))
        versions = Models.ProjectVersion.objects
        additional_values = []
        if settings.HAS_ORG_MANAGEMENT:
            additional_values = ["project__owner__name"]
            versions = versions.filter(
                Q(project__owner__id=obj.project.owner_id)
                | Q(project__owner__name=DEFAULT_HUB)
            )
        versions = versions.filter(reduce(or_, filters)).values(
            *additional_values, "project__name", "name", "kind", "state"
        )
        models = []
        artifacts = []
        components = []

        for v in versions:
            k = v.pop("kind")
            v["project"] = v.pop("project__name")
            v["owner"] = v.pop("project__owner__name", "default")
            if k == V1ProjectVersionKind.ARTIFACT:
                artifacts.append(v)
            elif k == V1ProjectVersionKind.MODEL:
                models.append(v)
            elif k == V1ProjectVersionKind.COMPONENT:
                components.append(v)

        namespace = (
            obj.namespace if hasattr(obj, "namespace") else conf.get(K8S_NAMESPACE)
        )
        agent = None
        if hasattr(obj, "agent") and obj.agent:
            agent = {"name": obj.agent.name, "version": obj.agent.version}
        queue = None
        if hasattr(obj, "queue") and obj.queue:
            queue = {"name": obj.queue.name}
        artifacts_store = None
        if hasattr(obj, "artifacts_store") and obj.artifacts_store:
            artifacts_store = {"name": obj.artifacts_store.name}
        tensorboard = None
        if tensorboard:
            tensorboard = {
                "name": tensorboard.name,
                "status": tensorboard.status,
                "uuid": tensorboard.uuid.hex,
            }
        build = None
        if build:
            build = {
                "name": build.name,
                "status": build.status,
                "uuid": build.uuid.hex,
            }
        component = None
        if obj.component_state:
            component = {
                "state": obj.component_state.hex,
                "versions": components,
                "count": Models.Run.objects.filter(
                    project_id=obj.project_id,
                    component_state=obj.component_state,
                ).count(),
            }
        return {
            "namespace": namespace,
            "agent": agent,
            "queue": queue,
            "artifacts_store": artifacts_store,
            "tensorboard": tensorboard,
            "build": build,
            "component": component,
            "models": models,
            "artifacts": artifacts,
        }



STATUS_UPDATE_COLUMNS_ONLY = [
    "id",
    "status",
    "status_conditions",
    "started_at",
    "updated_at",
    "finished_at",
    "duration",
    "wait_time",
    "meta_info",
]
STATUS_UPDATE_COLUMNS_DEFER = [
    "original",
    "cloning_kind",
    "description",
    "inputs",
    "outputs",
    "tags",
    "description",
    "live_state",
    "readme",
    "content",
    "managed_by",
    "pending",
    "controller",
    "pipeline",
    "state",
]
DEFAULT_COLUMNS_DEFER = ["description", "readme", "content", "raw_content"]



class SchedulerCeleryTasks:
    """Scheduler celery tasks.

    N.B. make sure that the task name is not < 128.
    """

    SCHEDULER_HEALTH = "scheduler_health"

    RUNS_PREPARE = "runs_prepare"
    RUNS_START = "runs_start"
    RUNS_BUILT = "runs_built"
    RUNS_STOP = "runs_stop"
    RUNS_DELETE = "runs_delete"
    RUNS_SET_ARTIFACTS = "runs_set_artifacts"
    RUNS_HOOKS = "runs_hooks"
    RUNS_START_IMMEDIATELY = "runs_start_immediately"
    RUNS_NOTIFY_STATUS = "runs_notify_status"
    RUNS_NOTIFY_DONE = "runs_notify_done"
    RUNS_CHECK_PIPELINE = "runs_check_pipeline"
    RUNS_CHECK_ORPHAN_PIPELINE = "runs_check_orphan_pipeline"
    RUNS_CHECK_EARLY_STOPPING = "runs_check_early_stopping"
    RUNS_WAKEUP_SCHEDULE = "runs_wakeup_schedule"
    RUNS_ITERATE = "runs_iterate"
    RUNS_TUNE = "runs_tune"
    DELETE_ARCHIVED_PROJECT = "delete_archived_project"
    DELETE_ARCHIVED_RUN = "delete_archived_run"
    STATS_CALCULATION_PROJECT = "stats_calculation_project"


class CronsCeleryTasks:
    """Crons celery tasks.

    N.B. make sure that the task name is not < 128.
    """

    CRONS_HEALTH = "crons_health"

    DELETE_ARCHIVED_PROJECTS = "delete_archived_projects"
    DELETE_IN_PROGRESS_PROJECTS = "delete_in_progress_projects"
    DELETE_ARCHIVED_RUNS = "delete_archived_runs"
    DELETE_IN_PROGRESS_RUNS = "delete_in_progress_runs"

    STATS_CALCULATION_PROJECTS = "stats_calculation_projects"

    HEARTBEAT_RUNS = "heartbeat_runs"
    HEARTBEAT_OUT_OF_SYNC_SCHEDULES = "heartbeat_out_of_sync_schedules"
    HEARTBEAT_STOPPING_RUNS = "heartbeat_stopping_runs"
    HEARTBEAT_PROJECT_LAST_UPDATED = "heartbeat_project_last_updated"



ENABLED = "ENABLED"
URL = "URL"
GLOBAL_COUNTDOWN = "GLOBAL_COUNTDOWN"
DEFAULT_PREFIX = "DEFAULT_PREFIX"
TOKEN = "TOKEN"
CONTAINER = "CONTAINER"
BOOKMARKS = "BOOKMARKS"
ARCHIVES = "ARCHIVES"
DELETION = "DELETION"
SEARCHES = "SEARCHES"
IMAGE_PULL_POLICY = "IMAGE_PULL_POLICY"
DEFAULT_DOMAIN = "DEFAULT_DOMAIN"
NOTIFICATIONS = "NOTIFICATIONS"
ORGANIZATIONS = "ORGANIZATIONS"
ACTIVITY_LOGS = "ACTIVITY_LOGS"
HUB = "HUB"
